/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "intc.hh"
#include "superh.hh"

RegistresIntc::RegistresIntc(void) : Memoire(0x1FF, ONCHIP) {
}

unsigned short RegistresIntc::getIPRA(void) {
  return getMot(0xE2);
}

unsigned short RegistresIntc::getIPRB(void) {
  return getMot(0x60);
}

unsigned short RegistresIntc::getICR(void) {
  return getMot(0xE0);
}

ostream& operator<<(ostream& os, RegistresIntc& ri) {
  char chaine[100];
  sprintf(chaine, "IPRA\t: %10X\n", ri.getIPRA());
  os << chaine;
  sprintf(chaine, "IPRB\t: %10X\n", ri.getIPRB());
  os << chaine;
  sprintf(chaine, "ICR\t: %10X\n", ri.getICR());
  os << chaine;
  return os;
}

RegistresIntc Intc::memoire;
FileMessage Intc::fm(INTC_MSG);
const int Intc::INT_IRL[15] = {17, 16, 15, 14, 13, 12, 11, 10,
                               9, 8, 7, 6, 5, 4, 3};

/*
Intc::Intc(void) {
  struct sigaction action;

  action.sa_handler = &gererInterruption;
  sigaction(SIGUSR1, &action, NULL);
  sigaction(SIGUSR2, &action, NULL);
  
  sigset_t masque;
  sigemptyset(&masque);
  sigsuspend(&masque);
}
*/

void Intc::lancer(void) {
  stop = false;
  Message mess;
  while (!stop) {
    fm.recevoirBloquant(mess, 0);
    fm.envoyer(mess);
    while (fm.nombreMessages() > 0) executer();
  }
}

void Intc::executer(void) {
  SuperH *proc = new(SH2) SuperH();
  Message interruption;
  // On commence par regarder quelles interruptions ont t reues
  // sur la file de message et on choisi celle de priorit la plus
  // leve

  if (fm.recevoir(interruption, INTC_STOP)) stop = true;
  
  if (fm.recevoir(interruption, INT_NMI)) {
    proc->vecteur = 0xB;
    proc->niveau = 0x10;
    proc->interruption = true;
  }
  else {
    if (proc->SR.partie.I < 0xF) {
      if (fm.recevoir(interruption, INT_USER_BREAK)) {
        cerr << "USER BREAK\n";
        proc->vecteur = 0xC;
        proc->niveau = 0xF;
        proc->interruption = true;
      }
      else {
        // FIXME pour l'instant les interruptions des modules on-chip ne
        // sont pas gres
	/* 
        int pDIVU = Intc::memoire.getIPRA() >> 12;
        int pDMAC = (Intc::memoire.getIPRA() >> 8) & 0x0F;
        int pWDT = (Intc::memoire.getIPRA() >> 4) & 0x00F;
        int pSCI = Intc::memoire.getIPRB() >> 12;
        int pFRT = (Intc::memoire.getIPRB() >> 8) & 0x0F;
	*/

	unsigned int i = 15;
	bool een = true;
	while((i > 0) && een) {
	  if (fm.recevoir(interruption, INT_IRL[i - 1])) {
	    sscanf(interruption.getTexte(), "%c", &proc->vecteur);
	    proc->niveau = i;
	    proc->interruption = true;
	    een = false;
	  }
	  i--;
	}
      }
    }
  }
  delete proc;
}
