/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "superh.hh"
#include "message.hh"
#include "sem.hh"

SuperH::SuperH(void) {
}

SuperH::SuperH(Memoire& mem) {
  memoire = &mem;
  SR.partie.T = SR.partie.S = SR.partie.Q = SR.partie.M = 0;
  SR.partie.I = 0xF;
  SR.partie.inutile1 = 0;
  SR.partie.inutile2 = 0;
  VBR = 0;

  PC = memoire->getGrosmot(0) + 4;

  R[15] = memoire->getGrosmot(4);

  stop = false;
  interruption = false;
  _delai = 0;
}

void SuperH::lancer(void) {
  unsigned int compte = 0;
  
  while (!stop) {
    //if (!compte) kill(Reveil<Intc>::pid(), SIGUSR1);
    if (_delai) {
      unsigned long tmp = PC;
      PC = _delai;
      executer(memoire->getMot(PC - 4));
      PC = tmp;
      _delai = 0;
    }
    else {
      if (interruption) {
	interruption = false;
        R[15] -= 4;
	memoire->setGrosmot(R[15], SR.tout);
	R[15] -= 4;
	memoire->setGrosmot(R[15], PC);
	SR.partie.I = niveau;
	PC = memoire->getGrosmot(VBR + (vecteur << 2)) + 4;
      }
      executer(memoire->getMot(PC - 4));
    }
    compte++;
    compte %= 10000;
  }
}

ostream& operator<<(ostream& os, const SuperH& sh) {
  char chaine[100];
  for(int j = 0;j < 4;j++) {
    os << "\t";
    for(int i = 4*j;i < 4 * (j + 1);i++) {
      sprintf(chaine, "R%-4d%10X  ",i,sh.R[i]);
      os << chaine;
    }
    os << "\n";
  }

  sprintf(chaine, "\tSR    M  %-11dQ  %-11dI  %-11XS  %-11dT  %d\n",
	  sh.SR.partie.M,sh.SR.partie.Q,sh.SR.partie.I,
	  sh.SR.partie.S,sh.SR.partie.T);
  os << chaine;

  sprintf(chaine, "\tGBR\t%10X\tMACH\t%10X\tPR\t%10X\n",sh.GBR,sh.MACH,sh.PR);
  os << chaine;

  sprintf(chaine, "\tVBR\t%10X\tMACL\t%10X\tPC\t%10X\n",sh.VBR,sh.MACL,sh.PC);
  os << chaine;

  return os;
}

// addition de deux registres source + dest -> dest
void SuperH::add(long source, long dest) {
  R[dest] += R[source];
  PC += 2;
}

// addition d'un entier sur 8 bits et d'un registre source + dest -> dest
void SuperH::addi(long source, long dest) {
  if (source & 128)                           // si l'entier est ngatif
    R[dest] += (0xFFFFFF00 | (long) source);  // on complte par des '1'
  else                                        // sinon
    R[dest] += (0x000000FF & (long) source);  // par des '0'
  PC += 2;
}

// addition de deux registres avec retenue source + dest -> dest
void SuperH::addc(long source, long dest) {
  unsigned long tmp0, tmp1;

  tmp1 = R[source] + R[dest];
  tmp0 = R[dest];

  R[dest] = tmp1 + SR.partie.T;
  if (tmp0 > tmp1)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  if (tmp1 > R[dest])
    SR.partie.T = 1;
  PC += 2;
}

/* addition de deux registres avec vrification du dbordement
   source + dest -> dest */
void SuperH::addv(long source, long dest) {
  long i,j,k;

  if ((long) R[dest] >= 0) i = 0; else i = 1;
  //i = ((long) R[dest] < 0);
  
  if ((long) R[source] >= 0) j = 0; else j = 1;
  //j = ((long) R[source] < 0);
  
  j += i;
  R[dest] += R[source];

  if ((long) R[dest] >= 0) k = 0; else k = 1;
  //k = ((long) R[dest] < 0);

  k += j;
  
  if (j == 0 || j == 2)
    if (k == 1) SR.partie.T = 1;
    else SR.partie.T = 0;
  else SR.partie.T = 0;
  //SR.partie.T = ((j == 0 || j == 2) && (k == 1));
  PC += 2;
}

void SuperH::y_and(long source, long dest) {
  R[dest] &= R[source];
  PC += 2;
}

void SuperH::andi(long source) {
  R[0] &= (0x000000FF & (long) source);
  PC += 2;
}

void SuperH::andm(long source) {
  long temp;

  temp = (long) memoire->getOctet(GBR + R[0]);
  temp &= (0x000000FF & (long) source);
  memoire->setOctet((GBR + R[0]),temp);
  PC += 2;
}
 
void SuperH::bf(long d) { // FIXME peut etre amelior
  long disp;

  if ((d&0x80)==0)
    disp=(0x000000FF & (long)d);
  else
    disp=(0xFFFFFF00 | (long)d);
  if (SR.partie.T == 0)
    PC = PC+(disp<<1)+4;
  else PC+=2;
}

void SuperH::bfs(long d) { // FIXME peut tre amlior
  long disp;
  unsigned long temp;

  temp = PC;
  if ((d & 0x80) == 0)
    disp = (0x000000FF & (long) d);
  else
    disp = (0xFFFFFF00 | (long) d);
  if (SR.partie.T == 0) {
    PC = PC + (disp << 1) + 4;

    _delai = temp + 2;
  }
  else {
    PC += 2;

  }
}

void SuperH::bra(long d) {
  unsigned long temp;

  long disp;
  if ((d&0x800)==0) disp=(0x00000FFF & d);
  else disp=(0xFFFFF000 | d);
  temp=PC;
  PC = PC + (disp<<1) + 4;

  _delai = temp + 2;
}

void SuperH::braf(long m) {
  unsigned long temp;

  temp = PC;
  PC += R[m] + 4; /* FIXME : Correction trouve dans semu, est-ce bon ?
		     Hyperion est d'accord ! */

  _delai = temp + 2;
}

void SuperH::bsr(long d) {
  long disp;

  if ((d&0x800)==0) disp = (0x00000FFF & d);
  else disp = (0xFFFFF000 | d);
  PR = PC;
  PC = PC+(disp<<1) + 4;

  _delai = PR + 2;
}

void SuperH::bsrf(long m) {
  PR = PC;
  PC += R[m] + 4; // correction donne par Hyperion, merci
  _delai = PR + 2;
}

void SuperH::bt(long d) { // FIXME ya plus rapide
  long disp;

  if ((d&0x80)==0)
    disp=(0x000000FF & (long)d);
  else
    disp=(0xFFFFFF00 | (long)d);
  if (SR.partie.T == 1)
    PC = PC+(disp<<1)+4;
  else
    PC += 2;
}

void SuperH::bts(long d) {
  long disp;
  unsigned long temp;
  
  if (SR.partie.T) {
    temp = PC;
    if ((d & 0x80) == 0) disp = (0x000000FF & (long) d);
    else disp = (0xFFFFFF00 | (long)d);
    PC += (disp << 1) + 4;
    _delai = temp + 2;
  }
  else PC+=2;
}

void SuperH::clrmac(void) {
  MACH = 0;
  MACL = 0;
  PC += 2;
}

void SuperH::clrt(void) {
  SR.partie.T = 0;
  PC += 2;
}

void SuperH::cmpeq(long m, long n) {
  if (R[n] == R[m])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC += 2;
}

void SuperH::cmpge(long m, long n) {
  if ((long)R[n] >= (long)R[m])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC += 2;
}

void SuperH::cmpgt(long m, long n) {
  if ((long)R[n]>(long)R[m])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmphi(long m, long n) {
  if ((unsigned long)R[n]>(unsigned long)R[m])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmphs(long m, long n) {
  if ((unsigned long)R[n]>=(unsigned long)R[m])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmpim(long i) {
  long imm;

  if ((i&0x80)==0)
    imm = (0x000000FF & (long) i);
  else
    imm=(0xFFFFFF00 | (long) i);
  if (R[0] == (unsigned long) imm) // FIXME: ouais a doit tre bon...
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmppl(long n) {
  if ((long)R[n]>0)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmppz(long n) {
  if ((long)R[n]>=0)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmpstr(long m,long n) {
  unsigned long temp;
  long HH,HL,LH,LL;
  
  temp=R[n]^R[m];
  HH = (temp>>12) & 0x000000FF;
  HL = (temp>>8) & 0x000000FF;
  LH = (temp>>4) & 0x000000FF;
  LL = temp & 0x000000FF;
  HH = HH && HL && LH && LL;
  if (HH == 0)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::div0s(long m,long n) {
  if ((R[n]&0x80000000)==0)
    SR.partie.Q = 0;
  else
    SR.partie.Q = 1;
  if ((R[m]&0x80000000)==0)
    SR.partie.M = 0;
  else
    SR.partie.M = 1;
  SR.partie.T = !(SR.partie.M == SR.partie.Q);
  PC += 2;
}

void SuperH::div0u(void) {
  SR.partie.M = SR.partie.Q = SR.partie.T = 0;
  PC += 2;
}

void SuperH::div1(long m,long n) {
  unsigned long tmp0;
  unsigned char old_q, tmp1;
  
  old_q = SR.partie.Q;
  SR.partie.Q = (unsigned char)((0x80000000 & R[n])!=0);
  R[n] <<= 1;
  R[n]|=(unsigned long) SR.partie.T;
  switch(old_q){
  case 0:
    switch(SR.partie.M){
    case 0:
      tmp0 = R[n];
      R[n] -= R[m];
      tmp1 = (R[n] > tmp0);
      switch(SR.partie.Q){
      case 0:
	SR.partie.Q = tmp1;
	break;
      case 1:
	SR.partie.Q = (unsigned char) (tmp1 == 0);
	break;
      }
      break;
    case 1:
      tmp0 = R[n];
      R[n] += R[m];
      tmp1 = (R[n] < tmp0);
      switch(SR.partie.Q){
      case 0:
	SR.partie.Q = (unsigned char) (tmp1 == 0);
	break;
      case 1:
	SR.partie.Q = tmp1;
	break;
      }
      break;
    }
    break;
  case 1:switch(SR.partie.M){
  case 0:
    tmp0 = R[n];
    R[n] += R[m];
    tmp1 = (R[n] < tmp0);
    switch(SR.partie.Q){
    case 0:
      SR.partie.Q = tmp1;
      break;
    case 1:
      SR.partie.Q = (unsigned char) (tmp1 == 0);
      break;
    }
    break;
  case 1:
    tmp0 = R[n];
    R[n] -= R[m];
    tmp1 = (R[n] > tmp0);
    switch(SR.partie.Q){
    case 0:
      SR.partie.Q = (unsigned char) (tmp1 == 0);
      break;
    case 1:
      SR.partie.Q = tmp1;
      break;
    }
    break;
  }
  break;
  }
  SR.partie.T = (SR.partie.Q == SR.partie.M);
  PC += 2;
}


void SuperH::dmuls(long m,long n) {
  unsigned long RnL,RnH,RmL,RmH,Res0,Res1,Res2;
  unsigned long temp0,temp1,temp2,temp3;
  long tempm,tempn,fnLmL;
  
  tempn = (long)R[n];
  tempm = (long)R[m];
  if (tempn < 0) tempn = 0 - tempn;
  if (tempm < 0) tempm = 0 - tempm;
  if ((long) (R[n] ^ R[m]) < 0) fnLmL = -1;
  else fnLmL = 0;
  
  temp1 = (unsigned long) tempn;
  temp2 = (unsigned long) tempm;

  RnL = temp1 & 0x0000FFFF;
  RnH = (temp1 >> 16) & 0x0000FFFF;
  RmL = temp2 & 0x0000FFFF;
  RmH = (temp2 >> 16) & 0x0000FFFF;
  
  temp0 = RmL * RnL;
  temp1 = RmH * RnL;
  temp2 = RmL * RnH;
  temp3 = RmH * RnH;

  Res2 = 0;
  Res1 = temp1 + temp2;
  if (Res1 < temp1) Res2 += 0x00010000;

  temp1 = (Res1 << 16) & 0xFFFF0000;
  Res0 = temp0 + temp1;
  if (Res0 < temp0) Res2++;
  
  Res2 = Res2 + ((Res1 >> 16) & 0x0000FFFF) + temp3;

  if (fnLmL < 0) {
    Res2 = ~Res2;
    if (Res0 == 0)
      Res2++;
    else
      Res0 =(~Res0) + 1;
  }
  MACH = Res2;
  MACL = Res0;
  PC += 2;
}

void SuperH::dmulu(long m,long n) {
  unsigned long RnL,RnH,RmL,RmH,Res0,Res1,Res2;
  unsigned long temp0,temp1,temp2,temp3;

  RnL = R[n] & 0x0000FFFF;
  RnH = (R[n] >> 16) & 0x0000FFFF;
  RmL = R[m] & 0x0000FFFF;
  RmH = (R[m] >> 16) & 0x0000FFFF;

  temp0 = RmL * RnL;
  temp1 = RmH * RnL;
  temp2 = RmL * RnH;
  temp3 = RmH * RnH;
  
  Res2 = 0;
  Res1 = temp1 + temp2;
  if (Res1 < temp1) Res2 += 0x00010000;
  
  temp1 = (Res1 << 16) & 0xFFFF0000;
  Res0 = temp0 + temp1;
  if (Res0 < temp0) Res2++;
  
  Res2 = Res2 + ((Res1 >> 16) & 0x0000FFFF) + temp3;

  MACH = Res2;
  MACL = Res0;
  PC += 2;
}

void SuperH::dt(long n) {
  R[n]--;
  if (R[n] == 0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC += 2;
}

void SuperH::extsb(long m,long n) {
  R[n] = R[m];
  if ((R[m]&0x00000080)==0) R[n]&=0x000000FF;
  else R[n]|=0xFFFFFF00;
  PC += 2;
}

void SuperH::extsw(long m,long n) {
  R[n] = R[m];
  if ((R[m]&0x00008000)==0) R[n]&=0x0000FFFF;
  else R[n]|=0xFFFF0000;
  PC += 2;
}

void SuperH::extub(long m,long n) {
  R[n] = R[m];
  R[n] &= 0x000000FF;
  PC += 2;
}

void SuperH::extuw(long m,long n) {
  R[n] = R[m];
  R[n] &= 0x0000FFFF;
  PC += 2;
}

void SuperH::jmp(long m) {
  unsigned long temp;

  temp=PC;
  PC = R[m] + 4;
  _delai = temp + 2;
}

void SuperH::jsr(long m) {
  PR = PC;
  PC = R[m] + 4;
  _delai = PR + 2;
}

void SuperH::ldcgbr(long m) {
  GBR = R[m];
  PC += 2;
}

void SuperH::ldcmgbr(long m) {
  GBR = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldcmsr(long m) {
  SR.tout = memoire->getGrosmot(R[m]) & 0x000003F3;
  R[m] += 4;
  PC += 2;
}

void SuperH::ldcmvbr(long m) {
  VBR = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldcsr(long m) {
  SR.tout = R[m]&0x000003F3;
  PC += 2;
}

void SuperH::ldcvbr(long m) {
  VBR = R[m];
  PC += 2;
}

void SuperH::ldsmach(long m) {
  MACH = R[m];
  PC+=2;
}

void SuperH::ldsmacl(long m) {
  MACL = R[m];
  PC += 2;
}

void SuperH::ldsmmach(long m) {
  MACH = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldsmmacl(long m) {
  MACL = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldsmpr(long m) {
  PR = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldspr(long m) {
  PR = R[m];
  PC += 2;
}

void SuperH::macl(long m,long n) {
  unsigned long RnL,RnH,RmL,RmH,Res0,Res1,Res2;
  unsigned long temp0,temp1,temp2,temp3;
  long tempm,tempn,fnLmL;
  
  tempn = (long) memoire->getGrosmot(R[n]);
  R[n] += 4;
  tempm = (long) memoire->getGrosmot(R[m]);
  R[m] += 4;


  if ((long) (tempn^tempm) < 0) fnLmL =- 1;
  else fnLmL = 0;
  if (tempn < 0) tempn = 0 - tempn;
  if (tempm < 0) tempm = 0 - tempm;

  temp1 = (unsigned long) tempn;
  temp2 = (unsigned long) tempm;

  RnL = temp1 & 0x0000FFFF;
  RnH = (temp1 >> 16) & 0x0000FFFF;
  RmL = temp2 & 0x0000FFFF;
  RmH = (temp2 >> 16) & 0x0000FFFF;

  temp0 = RmL * RnL;
  temp1 = RmH * RnL;
  temp2 = RmL * RnH;
  temp3 = RmH * RnH;
     
  Res2 = 0;
  Res1 = temp1 + temp2;
  if (Res1 < temp1) Res2 += 0x00010000;

  temp1 = (Res1 << 16) & 0xFFFF0000;
  Res0 = temp0 + temp1;
  if (Res0 < temp0) Res2++;

  Res2=Res2+((Res1>>16)&0x0000FFFF)+temp3;

  if(fnLmL < 0){
    Res2=~Res2;
    if (Res0==0) Res2++;
    else Res0=(~Res0)+1;
  }
  if(SR.partie.S == 1){
    Res0=MACL+Res0;
    if (MACL>Res0) Res2++;
    Res2+=(MACH&0x0000FFFF);

    if(((long)Res2<0)&&(Res2<0xFFFF8000)){
      Res2=0x00008000;
      Res0=0x00000000;
    }
    if(((long)Res2>0)&&(Res2>0x00007FFF)){
      Res2=0x00007FFF;
      Res0=0xFFFFFFFF;
    };
    
    MACH=Res2;
    MACL=Res0;
  }
  else {
    Res0=MACL+Res0;
    if (MACL>Res0) Res2++;
    Res2+=MACH;

    MACH=Res2;
    MACL=Res0;
  }
  PC+=2;
}

void SuperH::macw(long m,long n) {
  long tempm,tempn,dest,src,ans;
  unsigned long templ;
  tempn=(long) memoire->getMot(R[n]);
  R[n]+=2;
  tempm=(long) memoire->getMot(R[m]);
  R[m]+=2;
  templ=MACL;
  tempm=((long)(short)tempn*(long)(short)tempm);

  if ((long)MACL>=0) dest=0;
  else dest=1;
  if ((long)tempm>=0) {
    src=0;
    tempn=0;
  }
  else {
    src=1;
    tempn=0xFFFFFFFF;
  }
  src+=dest;
  MACL+=tempm;
  if ((long)MACL>=0) ans=0;
  else ans=1;
  ans+=dest;
  if (SR.partie.S == 1) {
    if (ans==1) {
      if (src==0) MACL=0x7FFFFFFF;
      if (src==2) MACL=0x80000000;
    }
  }
  else {
    MACH+=tempn;
    if (templ>MACL) MACH+=1;
  }
  PC+=2;
}

void SuperH::mov(long m,long n) {
  R[n]=R[m];
  PC+=2;
}

void SuperH::mova(long d) {
  long disp;

  disp=(0x000000FF & (long)d);
  R[0]=(PC&0xFFFFFFFC)+(disp<<2);
  PC+=2;
}

void SuperH::movbl(long m,long n) {
  R[n] = (long) memoire->getOctet(R[m]);
  if ((R[n] & 0x80) == 0)
    R[n] &= 0x000000FF;
  else
    R[n] |= 0xFFFFFF00;
  PC += 2;
}

void SuperH::movbl0(long m,long n) {
  R[n]=(long) memoire->getOctet(R[m] + R[0]);
  if ((R[n]&0x80)==0)
    R[n] &= 0x000000FF;
  else
    R[n] |= 0xFFFFFF00;
  PC += 2;
}

void SuperH::movbl4(long m,long d) {
  long disp;

  disp=(0x0000000F & (long)d);
  R[0] = memoire->getOctet(R[m] + disp);
  if ((R[0]&0x80)==0) R[0]&=0x000000FF;
  else R[0]|=0xFFFFFF00;
  PC+=2;
}

void SuperH::movblg(long d) {
  long disp;
  
  disp=(0x000000FF & (long)d);
  R[0]=(long) memoire->getOctet(GBR + disp);
  if ((R[0]&0x80)==0) R[0]&=0x000000FF;
  else R[0]|=0xFFFFFF00;
  PC+=2;
}

void SuperH::movbm(long m,long n) {
  memoire->setOctet((R[n] - 1),R[m]);
  R[n] -= 1;
  PC += 2;
}

void SuperH::movbp(long m,long n) {
  R[n] = (long) memoire->getOctet(R[m]);
  if ((R[n]&0x80)==0)
    R[n] &= 0x000000FF;
  else
    R[n] |= 0xFFFFFF00;
  if (n != m)
    R[m] += 1;
  PC += 2;
}

void SuperH::movbs(long m,long n) {
  memoire->setOctet(R[n],R[m]);
  PC += 2;
}

void SuperH::movbs0(long m,long n) {
  memoire->setOctet(R[n] + R[0], R[m]);
  PC += 2;
}

void SuperH::movbs4(long d,long n) {
  long disp;

  disp=(0x0000000F & (long)d);
  memoire->setOctet(R[n]+disp,R[0]);
  PC+=2;
}

void SuperH::movbsg(long d) {
  long disp;

  disp=(0x000000FF & (long)d);
  memoire->setOctet(GBR + disp,R[0]);
  PC += 2;
}

void SuperH::movi(long i,long n) {
  if ((i & 0x80) == 0)
    R[n] = (0x000000FF & (long)i);
  else
    R[n] = (0xFFFFFF00 | (long)i);
  PC += 2;
}

void SuperH::movli(long d,long n) {
  long disp;

  disp=(0x000000FF & (long)d);
  R[n] = memoire->getGrosmot((PC & 0xFFFFFFFC) + (disp << 2));
  PC += 2;
}

void SuperH::movll(long m,long n) {
  R[n] = memoire->getGrosmot(R[m]);
  PC += 2;
}

void SuperH::movll0(long m,long n) {
  R[n] = memoire->getGrosmot(R[m] + R[0]);
  PC += 2;
}

void SuperH::movll4(long m,long d,long n) {
  long disp;

  disp=(0x0000000F & (long)d);
  R[n] = memoire->getGrosmot(R[m] + (disp << 2));
  PC += 2;
}

void SuperH::movllg(long d) {
  long disp;

  disp=(0x000000FF & (long)d);
  R[0] = memoire->getGrosmot(GBR + (disp << 2));
  PC+=2;
}

void SuperH::movlm(long m,long n) {
  memoire->setGrosmot(R[n] - 4,R[m]);
  R[n] -= 4;
  PC += 2;
}

void SuperH::movlp(long m,long n) {
  R[n] = memoire->getGrosmot(R[m]);
  if (n != m) R[m] += 4;
  PC += 2;
}

void SuperH::movls(long m,long n) {
  memoire->setGrosmot(R[n], R[m]);
  PC += 2;
}

void SuperH::movls0(long m,long n) {
  memoire->setGrosmot(R[n] + R[0], R[m]);
  PC += 2;
}

void SuperH::movls4(long m,long d,long n) {
  long disp;

  disp = (0x0000000F & (long) d);
  memoire->setGrosmot(R[n]+(disp<<2),R[m]);
  PC += 2;
}

void SuperH::movlsg(long d) {
  long disp;

  disp=(0x000000FF & (long)d);
  memoire->setGrosmot(GBR+(disp<<2),R[0]);
  PC+=2;
}

void SuperH::movt(long n) {
  R[n] = (0x00000001 & SR.tout);
  PC += 2;
}

void SuperH::movwi(long d,long n) {
  long disp;

  disp=(0x000000FF & (long)d);
  R[n]=(long) memoire->getMot(PC + (disp<<1));
  if ((R[n]&0x8000)==0) R[n]&=0x0000FFFF;
  else R[n]|=0xFFFF0000;
  PC+=2;
}

void SuperH::movwl(long m,long n) {
  R[n]=(long) memoire->getMot(R[m]);
  if ((R[n] & 0x8000) == 0)
    R[n] &= 0x0000FFFF;
  else
    R[n] |= 0xFFFF0000;
  PC += 2;
}

void SuperH::movwl0(long m,long n) {
  R[n]=(long) memoire->getMot(R[m]+R[0]);
  if ((R[n]&0x8000)==0)
    R[n] &= 0x0000FFFF;
  else
    R[n] |= 0xFFFF0000;
  PC+=2;
}

void SuperH::movwl4(long m,long d) {
  long disp;

  disp=(0x0000000F & (long)d);
  R[0] = memoire->getMot(R[m]+(disp<<1));
  if ((R[0]&0x8000)==0) R[0]&=0x0000FFFF;
  else R[0]|=0xFFFF0000;
  PC+=2;
}

void SuperH::movwlg(long d) {
  long disp;

  disp=(0x000000FF & (long)d);
  R[0]=(long) memoire->getMot(GBR+(disp<<1));
  if ((R[0]&0x8000)==0) R[0]&=0x0000FFFF;
  else R[0]|=0xFFFF0000;
  PC += 2;
}

void SuperH::movwm(long m,long n) {
  memoire->setMot(R[n] - 2,R[m]);
  R[n] -= 2;
  PC += 2;
}

void SuperH::movwp(long m,long n) {
  R[n] = (long) memoire->getMot(R[m]);
  if ((R[n]&0x8000)==0)
    R[n] &= 0x0000FFFF;
  else
    R[n] |= 0xFFFF0000;
  if (n != m)
    R[m] += 2;
  PC += 2;
}

void SuperH::movws(long m,long n) {
  memoire->setMot(R[n],R[m]);
  PC += 2;
}

void SuperH::movws0(long m,long n) {
  memoire->setMot(R[n] + R[0], R[m]);
  PC+=2;
}

void SuperH::movws4(long d,long n) {
  long disp;

  disp=(0x0000000F & (long)d);
  memoire->setMot(R[n]+(disp<<1),R[0]);
  PC+=2;
}

void SuperH::movwsg(long d) {
  long disp;

  disp=(0x000000FF & (long)d);
  memoire->setMot(GBR+(disp<<1),R[0]);
  PC+=2;
}

void SuperH::mull(long m,long n) {
  MACL = R[n] * R[m];
  PC+=2;
}

void SuperH::muls(long m,long n) {
  MACL = ((long)(short)R[n]*(long)(short)R[m]);
  PC += 2;
}

void SuperH::mulu(long m,long n) {
  MACL = ((unsigned long)(unsigned short)R[n]
	*(unsigned long)(unsigned short)R[m]);
  PC+=2;
}

void SuperH::neg(long m,long n) {
  R[n]=0-R[m];
  PC+=2;
}

void SuperH::negc(long m,long n) {
  unsigned long temp;
  
  temp=0-R[m];
  R[n] = temp - SR.partie.T;
  if (0 < temp) SR.partie.T=1;
  else SR.partie.T=0;
  if (temp < R[n]) SR.partie.T=1;
  PC+=2;
}

void SuperH::nop(void) {
  PC += 2;
}

void SuperH::y_not(long source, long dest) {
  R[dest] = ~R[source];
  PC += 2;
}

void SuperH::y_or(long source, long dest) {
  R[dest] |= R[source];
  PC += 2;
}

void SuperH::ori(long source) {
  R[0] |= (0x000000FF & (long) source);
  PC += 2;
}

void SuperH::orm(long source) {
  long temp;

  temp = (long) memoire->getOctet(GBR + R[0]);
  temp |= (0x000000FF & (long) source);
  memoire->setOctet(GBR + R[0],temp);
  PC += 2;
}

void SuperH::rotcl(long n) {
  long temp;

  if ((R[n]&0x80000000)==0) temp=0;
  else temp=1;
  R[n]<<=1;
  if (SR.partie.T == 1) R[n]|=0x00000001;
  else R[n]&=0xFFFFFFFE;
  if (temp==1) SR.partie.T=1;
  else SR.partie.T=0;
  PC+=2;
}

void SuperH::rotcr(long n) {
  long temp;

  if ((R[n]&0x00000001)==0) temp=0;
  else temp=1;
  R[n]>>=1;
  if (SR.partie.T == 1) R[n]|=0x80000000;
  else R[n]&=0x7FFFFFFF;
  if (temp==1) SR.partie.T=1;
  else SR.partie.T=0;
  PC+=2;
}

void SuperH::rotl(long n) {
  if ((R[n]&0x80000000)==0) SR.partie.T=0;
  else SR.partie.T=1;
  R[n]<<=1;
  if (SR.partie.T==1) R[n]|=0x00000001;
  else R[n]&=0xFFFFFFFE;
  PC+=2;
}

void SuperH::rotr(long n) {
  if ((R[n]&0x00000001)==0) SR.partie.T = 0;
  else SR.partie.T = 1;
  R[n]>>=1;
  if (SR.partie.T == 1) R[n]|=0x80000000;
  else R[n]&=0x7FFFFFFF;
  PC+=2;
}

void SuperH::rte() {
  unsigned long temp;

  temp=PC;
  PC = memoire->getGrosmot(R[15]); // + 4;
  R[15] += 4;
  SR.tout = memoire->getGrosmot(R[15]) & 0x000003F3;
  R[15] += 4;
  _delai = temp + 2;
}

void SuperH::rts() {
  unsigned long temp;
  
  temp = PC;
  PC = PR + 4;

  _delai = temp + 2;
}

void SuperH::sett(void) {
  SR.partie.T = 1;
  PC += 2;
}

void SuperH::shal(long n) {
  if ((R[n] & 0x80000000) == 0) SR.partie.T = 0;
  else SR.partie.T = 1;
  R[n] <<= 1;
  PC += 2;
}

void SuperH::shar(long n) {
  long temp;

  if ((R[n]&0x00000001)==0) SR.partie.T = 0;
  else SR.partie.T = 1;
  if ((R[n]&0x80000000)==0) temp = 0;
  else temp = 1;
  R[n] >>= 1;
  if (temp == 1) R[n] |= 0x80000000;
  else R[n] &= 0x7FFFFFFF;
  PC += 2;
}

void SuperH::shll(long n) {
  if ((R[n]&0x80000000)==0) SR.partie.T=0;
  else SR.partie.T=1;
  R[n]<<=1;
  PC+=2;
}

void SuperH::shll2(long n) {
  R[n] <<= 2;
  PC+=2;
}

void SuperH::shll8(long n) {
  R[n]<<=8;
  PC+=2;
}

void SuperH::shll16(long n) {
  R[n]<<=16;
  PC+=2;
}

void SuperH::shlr(long n) {
  if ((R[n]&0x00000001)==0) SR.partie.T=0;
  else SR.partie.T=1;
  R[n]>>=1;
  R[n]&=0x7FFFFFFF;
  PC+=2;
}

void SuperH::shlr2(long n) {
  R[n]>>=2;
  R[n]&=0x3FFFFFFF;
  PC+=2;
}

void SuperH::shlr8(long n) {
  R[n]>>=8;
  R[n]&=0x00FFFFFF;
  PC+=2;
}

void SuperH::shlr16(long n) {
  R[n]>>=16;
  R[n]&=0x0000FFFF;
  PC+=2;
}

void SuperH::stcgbr(long n) {
  R[n]=GBR;
  PC+=2;
}

void SuperH::stcmgbr(long n) {
  R[n]-=4;
  memoire->setGrosmot(R[n],GBR);
  PC+=2;
}

void SuperH::stcmsr(long n) {
  R[n]-=4;
  memoire->setGrosmot(R[n],SR.tout);
  PC+=2;
}

void SuperH::stcmvbr(long n) {
  R[n]-=4;
  memoire->setGrosmot(R[n],VBR);
  PC+=2;
}

void SuperH::stcsr(long n) {
  R[n] = SR.tout;
  PC+=2;
}

void SuperH::stcvbr(long n) {
  R[n]=VBR;
  PC+=2;
}

void SuperH::stsmach(long n) {
  R[n]=MACH;
  PC+=2;
}

void SuperH::stsmacl(long n) {
  R[n]=MACL;
  PC+=2;
}

void SuperH::stsmmach(long n) {
  R[n] -= 4;
  memoire->setGrosmot(R[n],MACH); 
  PC+=2;
}

void SuperH::stsmmacl(long n) {
  R[n] -= 4;
  memoire->setGrosmot(R[n],MACL);
  PC+=2;
}

void SuperH::stsmpr(long n) {
  R[n] -= 4;
  memoire->setGrosmot(R[n],PR);
  PC+=2;
}

void SuperH::stspr(long n) {
  R[n] = PR;
  PC+=2;
}

void SuperH::sub(long m,long n) {
  R[n]-=R[m];
  PC+=2;
}

void SuperH::subc(long m,long n) {
  unsigned long tmp0,tmp1;
  
  tmp1 = R[n] - R[m];
  tmp0 = R[n];
  R[n] = tmp1 - SR.partie.T;
  if (tmp0 < tmp1) SR.partie.T = 1;
  else SR.partie.T = 0;
  if (tmp1 < R[n]) SR.partie.T = 1;
  PC += 2;
}

void SuperH::subv(long m,long n) {
  long dest,src,ans;

  if ((long)R[n]>=0) dest=0;
  else dest=1;
  if ((long)R[m]>=0) src=0;
  else src=1;
  src+=dest;
  R[n]-=R[m];
  if ((long)R[n]>=0) ans=0;
  else ans=1;
  ans+=dest;
  if (src==1) {
    if (ans==1) SR.partie.T=1;
    else SR.partie.T=0;
  }
  else SR.partie.T=0;
  PC+=2;
}

void SuperH::swapb(long m,long n) {
  unsigned long temp0,temp1;

  temp0=R[m]&0xffff0000;
  temp1=(R[m]&0x000000ff)<<8;
  R[n]=(R[m]>>8)&0x000000ff;
  R[n]=R[n]|temp1|temp0;
  PC+=2;
}

void SuperH::swapw(long m,long n) {
  unsigned long temp;
  temp=(R[m]>>16)&0x0000FFFF;
  R[n]=R[m]<<16;
  R[n]|=temp;
  PC+=2;
}

void SuperH::tas(long n) {
  long temp;

  temp=(long) memoire->getOctet(R[n]);
  if (temp==0) SR.partie.T=1;
  else SR.partie.T=0;
  temp|=0x00000080;
  memoire->setOctet(R[n],temp);
  PC+=2;
}

void SuperH::trapa(long i) {
  long imm;

  imm=(0x000000FF & i);
  R[15]-=4;
  memoire->setGrosmot(R[15],SR.tout);
  R[15]-=4;
  memoire->setGrosmot(R[15],PC - 2);
  PC = memoire->getGrosmot(VBR+(imm<<2)) + 4;
}

void SuperH::tst(long m,long n) {
  if ((R[n]&R[m])==0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC += 2;
}

void SuperH::tsti(long i) {
  long temp;

  temp=R[0]&(0x000000FF & (long)i);
  if (temp==0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC+=2;
}

void SuperH::tstm(long i) {
  long temp;

  temp=(long) memoire->getOctet(GBR+R[0]);
  temp&=(0x000000FF & (long)i);
  if (temp==0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC+=2;
}

void SuperH::y_xor(long source,long dest) {
  R[dest] ^= R[source];
  PC += 2;
}

void SuperH::xori(long source) {
  R[0] ^= (0x000000FF & (long) source);
  PC += 2;
}

void SuperH::xorm(long source) {
  long temp;

  temp = (long) memoire->getOctet(GBR + R[0]);
  temp ^= (0x000000FF & (long) source);
  memoire->setOctet(GBR + R[0],temp);
  PC += 2;
}

void SuperH::xtrct(long m,long n) {
  unsigned long temp;

  temp=(R[m]<<16)&0xFFFF0000;
  R[n]=(R[n]>>16)&0x0000FFFF;
  R[n]|=temp;
  PC+=2;
}

/********************************* NON IMPLEMENTE **************/
void SuperH::sleep(void) { throw InstructionNonImplementee("sleep"); }
