/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "superh.hh"

void SuperH::executer(unsigned short instruction) {

  switch (Instruction::a(instruction)) {
  case 0:
    switch (Instruction::d(instruction)) {
    case 2:
      switch (Instruction::c(instruction)) {
      case 0: stcsr(Instruction::b(instruction)); break;
      case 1: stcgbr(Instruction::b(instruction)); break;
      case 2: stcvbr(Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);
      }
      break;
    case 3:
      switch (Instruction::c(instruction)) {
      case 0: bsrf(Instruction::b(instruction)); break;
      case 2: braf(Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);
      }
      break;
    case 4: movbs0(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 5: movws0(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 6: movls0(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 8:
      switch (Instruction::c(instruction)) {
      case 0: clrt(); break;
      case 1: sett(); break;
      case 2: clrmac(); break;
      default: throw InstructionIllegale(instruction);
      }
      break;
    case 9:
      switch (Instruction::c(instruction)) {
      case 0: nop(); break;
      case 1: div0u(); break;
      case 2: movt(Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);
      }
      break;
    case 10:
      switch (Instruction::c(instruction)) {
      case 0: stsmach(Instruction::b(instruction)); break;
      case 1: stsmacl(Instruction::b(instruction)); break;
      case 2: stspr(Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);
      }
      break;
    case 11:
      switch (Instruction::c(instruction)) {
      case 0: rts(); break;
      case 1: sleep(); break;
      case 2: rte(); break;
      default: throw InstructionIllegale(instruction);
      }
      break;
    case 12: movbl0(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 13: movwl0(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 14: movll0(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 15: macl(Instruction::c(instruction), Instruction::b(instruction)); break;
    default: throw InstructionIllegale(instruction);
    }
    break;
  case 1: movls4(Instruction::c(instruction),Instruction::d(instruction),Instruction::b(instruction)); break;
  case 2:
    switch (Instruction::d(instruction)) {
    case 0: movbs(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 1: movws(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 2: movls(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 4: movbm(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 5: movwm(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 6: movlm(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 7: div0s(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 8: tst(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 9: y_and(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 10: y_xor(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 11: y_or(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 12: cmpstr(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 13: xtrct(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 14: mulu(Instruction::c(instruction), Instruction::b(instruction)); break;
    case 15: muls(Instruction::c(instruction), Instruction::b(instruction)); break;
    default: throw InstructionIllegale(instruction);;
    }
    break;
  case 3:
    switch(Instruction::d(instruction)) {
    case 0:  cmpeq (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 2:  cmphs (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 3:  cmpge (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 4:  div1  (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 5:  dmulu (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 6:  cmphi (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 7:  cmpgt (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 8:  sub   (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 10: subc  (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 11: subv  (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 12: add   (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 13: dmuls (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 14: addc  (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 15: addv  (Instruction::c(instruction), Instruction::b(instruction)); break;
    default: throw InstructionIllegale(instruction);;
    }
    break;
  case 4:
    switch(Instruction::d(instruction)) {
    case 0:
      switch(Instruction::c(instruction)) {
      case 0: shll  (Instruction::b(instruction)); break;
      case 1: dt    (Instruction::b(instruction)); break;
      case 2: shal  (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 1:
      switch(Instruction::c(instruction)) {
      case 0: shlr     (Instruction::b(instruction)); break;
      case 1: cmppz    (Instruction::b(instruction)); break;
      case 2: shar     (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 2:
      switch(Instruction::c(instruction)) {
      case 0: stsmmach (Instruction::b(instruction)); break;
      case 1: stsmmacl (Instruction::b(instruction)); break;
      case 2: stsmpr   (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 3:
      switch(Instruction::c(instruction)) {
      case 0: stcmsr   (Instruction::b(instruction)); break;
      case 1: stcmgbr  (Instruction::b(instruction)); break;
      case 2: stcmvbr  (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 4:
      switch(Instruction::c(instruction)) {
      case 0: rotl     (Instruction::b(instruction)); break;
      case 2: rotcl    (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 5:
      switch(Instruction::c(instruction)) {
      case 0: rotr     (Instruction::b(instruction)); break;
      case 1: cmppl    (Instruction::b(instruction)); break;
      case 2: rotcr    (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 6:
      switch(Instruction::c(instruction)) {
      case 0: ldsmmach (Instruction::b(instruction)); break;
      case 1: ldsmmacl (Instruction::b(instruction)); break;
      case 2: ldsmpr   (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 7:
      switch(Instruction::c(instruction)) {
      case 0: ldcmsr   (Instruction::b(instruction)); break;
      case 1: ldcmgbr  (Instruction::b(instruction)); break;
      case 2: ldcmvbr  (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 8:
      switch(Instruction::c(instruction)) {
      case 0: shll2    (Instruction::b(instruction)); break;
      case 1: shll8    (Instruction::b(instruction)); break;
      case 2: shll16   (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 9:
      switch(Instruction::c(instruction)) {
      case 0: shlr2    (Instruction::b(instruction)); break;
      case 1: shlr8    (Instruction::b(instruction)); break;
      case 2: shlr16   (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 10:
      switch(Instruction::c(instruction)) {
      case 0: ldsmach  (Instruction::b(instruction)); break;
      case 1: ldsmacl  (Instruction::b(instruction)); break;
      case 2: ldspr    (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 11:
      switch(Instruction::c(instruction)) {
      case 0: jsr      (Instruction::b(instruction)); break;
      case 1: tas      (Instruction::b(instruction)); break;
      case 2: jmp      (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 14:
      switch(Instruction::c(instruction)) {
      case 0: ldcsr    (Instruction::b(instruction)); break;
      case 1: ldcgbr   (Instruction::b(instruction)); break;
      case 2: ldcvbr   (Instruction::b(instruction)); break;
      default: throw InstructionIllegale(instruction);;
      }
      break;
    case 15: macw      (Instruction::c(instruction), Instruction::b(instruction)); break;
    default: throw InstructionIllegale(instruction);;
    }
    break;
  case 5: movll4(Instruction::c(instruction),Instruction::d(instruction),Instruction::b(instruction)); break;
  case 6:
    switch (Instruction::d(instruction)) {
    case 0:  movbl (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 1:  movwl (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 2:  movll (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 3:  mov   (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 4:  movbp (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 5:  movwp (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 6:  movlp (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 7:  y_not   (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 8:  swapb (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 9:  swapw (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 10: negc  (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 11: neg   (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 12: extub (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 13: extuw (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 14: extsb (Instruction::c(instruction), Instruction::b(instruction)); break;
    case 15: extsw (Instruction::c(instruction), Instruction::b(instruction)); break;
    }
    break;
  case 7: addi(Instruction::cd(instruction),Instruction::b(instruction)); break;
  case 8:
    switch (Instruction::b(instruction)) {
    case 0:  movbs4 (Instruction::d(instruction), Instruction::c(instruction)); break;
    case 1:  movws4 (Instruction::d(instruction), Instruction::c(instruction)); break;
    case 4:  movbl4 (Instruction::c(instruction), Instruction::d(instruction)); break;
    case 5:  movwl4 (Instruction::c(instruction), Instruction::d(instruction)); break;
    case 8:  cmpim  (Instruction::cd(instruction)); break;
    case 9:  bt     (Instruction::cd(instruction)); break;
    case 11: bf     (Instruction::cd(instruction)); break;
    case 13: bts    (Instruction::cd(instruction)); break;
    case 15: bfs    (Instruction::cd(instruction)); break;
    default: throw InstructionIllegale(instruction);
    }
    break;
  case 9: movwi (Instruction::cd(instruction),Instruction::b(instruction)); break;
  case 10: bra (Instruction::bcd(instruction)); break;
  case 11: bsr (Instruction::bcd(instruction)); break;
  case 12:
    switch(Instruction::b(instruction)) {
    case 0:  movbsg (Instruction::cd(instruction)); break;
    case 1:  movwsg (Instruction::cd(instruction)); break;
    case 2:  movlsg (Instruction::cd(instruction)); break;
    case 3:  trapa  (Instruction::cd(instruction)); break;
    case 4:  movblg (Instruction::cd(instruction)); break;
    case 5:  movwlg (Instruction::cd(instruction)); break;
    case 6:  movllg (Instruction::cd(instruction)); break;
    case 7:  mova   (Instruction::cd(instruction)); break;
    case 8:  tsti   (Instruction::cd(instruction)); break;
    case 9:  andi   (Instruction::cd(instruction)); break;
    case 10: xori   (Instruction::cd(instruction)); break;
    case 11: ori    (Instruction::cd(instruction)); break;
    case 12: tstm   (Instruction::cd(instruction)); break;
    case 13: andm   (Instruction::cd(instruction)); break;
    case 14: xorm   (Instruction::cd(instruction)); break;
    case 15: orm    (Instruction::cd(instruction)); break;
    }
    break;
  case 13: movli (Instruction::cd(instruction), Instruction::b(instruction)); break;
  case 14: movi  (Instruction::cd(instruction), Instruction::b(instruction)); break;
  default: throw InstructionIllegale(instruction);
  }
}
