/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef SUPERH_HH
#define SUPERH_HH

#include "memoire.hh"
#include "objet.hh"
#include "reveil.hh"
#include "intc.hh"

class Instruction {
public:
  static inline unsigned long a   (unsigned long ul) { return ((ul & 0xF000) >> 12); }
  static inline unsigned long b   (unsigned long ul) { return ((ul & 0x0F00) >> 8); }
  static inline unsigned long c   (unsigned long ul) { return ((ul & 0x00F0) >> 4); }
  static inline unsigned long d   (unsigned long ul) { return (ul & 0x000F); }
  static inline unsigned long cd  (unsigned long ul) { return (ul & 0x00FF); }
  static inline unsigned long bcd (unsigned long ul) { return (ul & 0x0FFF); }
};

class Espion;
class Intc;

class SuperH : public Objet {
private:
  unsigned long R[16];

  union {
    struct {
      unsigned long T:1;
      unsigned long S:1;
      unsigned long inutile2:2;
      unsigned long I:4;
      unsigned long Q:1;
      unsigned long M:1;
      unsigned long inutile1:22;
    } partie;
    unsigned long tout;
  } SR;
  unsigned long GBR;
  unsigned long VBR;

  unsigned long MACH;
  unsigned long MACL;
  unsigned long PR;
  unsigned long PC;

  Memoire *memoire;

  bool stop;
  bool interruption;
  unsigned char niveau;
  unsigned char vecteur;
  unsigned long _delai;
public:
  friend class Espion;
  friend class Intc;
  friend int main(int, char**); //FIXME

  SuperH(void); // NE PAS UTILISER !!! DANGEREUX !!!
  SuperH(Memoire&);

  void executer(unsigned short);
  void lancer(void);
  friend ostream& operator<<(ostream&, const SuperH&);

  /**** A ******************************************************************/
  void add     (long,long); // addition de deux registres                   1
  void addi    (long,long); // addition d'un entier et d'un registre        2
  void addc    (long,long); // addition de deux registres avec retenue      3
  void addv    (long,long); // addition de deux registres avec vrification 4
  void y_and   (long,long); // fait le `et' logique entre deux registres    5
  void andi    (long);      // fait le `et' logique entre un entier et R0   6
  void andm    (long);      // fait le `et' logique entre un entier et R0   7
  /**** B ******************************************************************/
  void bf      (long);      // branchement si T == 0                        8
  void bfs     (long);      // branchement si T == 0 avec dlai             9
  void bra     (long);      // branchement inconditionnel (avec dlai)     10
  void braf    (long);      // bra inconditionnel loign (avec dlai)     11
  void bsr     (long);      // branchement vers un sous-programme          12
  void bsrf    (long);      // branchement vers un sous-programme loign  13
  void bt      (long);      // branchement si T == 1                       14
  void bts     (long);      // branchement si T == 1 avec dlai            15
  /**** C ******************************************************************/
  void clrmac  (void);      // vide les registres MACH et MACL             16
  void clrt    (void);      // place le bit T  0                          17
  void cmpeq   (long,long); // ==                                          18
  void cmpge   (long,long); // >= sign                                    19
  void cmpgt   (long,long); // > sign                                     20
  void cmphi   (long,long); // > non sign                                 21
  void cmphs   (long,long); // >= sign                                    22
  void cmpim   (long);      // ==                                          23
  void cmppl   (long);      // > 0                                         24
  void cmppz   (long);      // >= 0                                        25
  void cmpstr  (long,long); // ???                                         26
  /**** D ******************************************************************/
  void div0s   (long,long); // \                                           27
  void div0u   (void);      //  > division                                 28
  void div1    (long,long); // /                                           29
  void dmuls   (long,long); // multiplication                              30
  void dmulu   (long,long); // multiplication                              31
  void dt      (long);      // dcrement et test                           32
  /**** E ******************************************************************/
  void extsb   (long,long); // tend                                       33
  void extsw   (long,long); // tend aussi    :-)                          34
  void extub   (long,long); // tend toujours :-|                          35
  void extuw   (long,long); // tend encore   :-(                          36
  /**** J ******************************************************************/
  void jmp     (long);      // saut avec dlai                             37
  void jsr     (long);      // saut vers un sous-programme                 38
  /**** L ******************************************************************/
  void ldcgbr  (long);      // charge le registre gbr                      39
  void ldcmgbr (long);      //                                             40
  void ldcmsr  (long);      //                                             41
  void ldcmvbr (long);      //                                             42
  void ldcsr   (long);      // charge le registre sr                       43
  void ldcvbr  (long);      // charge le registre vbr                      44
  void ldsmach (long);      //                                             45
  void ldsmacl (long);      //                                             46
  void ldsmmach(long);      //                                             47
  void ldsmmacl(long);      //                                             48
  void ldsmpr  (long);      //                                             49
  void ldspr   (long);      //                                             50
  /**** M ******************************************************************/
  void macl    (long,long); //                                             51
  void macw    (long,long); //                                             53
  void mov     (long,long); //                                             54
  void mova    (long);      //                                             55
  void movbl   (long,long); //                                             56
  void movbl0  (long,long); //                                             57
  void movbl4  (long,long); //                                             58
  void movblg  (long);      //                                             59
  void movbm   (long,long); //                                             60
  void movbp   (long,long); //                                             61
  void movbs   (long,long); //                                             62
  void movbs0  (long,long); //                                             63
  void movbs4  (long,long); //                                             64
  void movbsg  (long);      //                                             65
  void movi    (long,long); //                                             66
  void movli   (long,long); //                                             67
  void movll   (long,long); //                                             68
  void movll0  (long,long); //                                             69
  void movll4  (long,long,long); //                                        70
  void movllg  (long);      //                                             71
  void movlm   (long,long); //                                             72
  void movlp   (long,long); //                                             73
  void movls   (long,long); //                                             74
  void movls0  (long,long); //                                             75
  void movls4  (long,long,long); //                                        76
  void movlsg  (long);      //                                             77
  void movt    (long);      //                                             78
  void movwi   (long,long); //                                             79
  void movwl   (long,long); //                                             80
  void movwl0  (long,long); //                                             81
  void movwl4  (long,long); //                                             82
  void movwlg  (long);      //                                             83
  void movwm   (long,long); //                                             84
  void movwp   (long,long); //                                             85
  void movws   (long,long); //                                             86
  void movws0  (long,long); //                                             87
  void movws4  (long,long); //                                             88
  void movwsg  (long);      //                                             89
  void mull    (long,long); //                                             90
  void muls    (long,long); //                                             91
  void mulu    (long,long); //                                             92
  /**** N ******************************************************************/
  void neg     (long,long); // ngation                                    93
  void negc    (long,long); // ngation avec retenue                       94
  void nop     (void);      // rien                                        95
  void y_not   (long,long); // complment logique                          96
  /**** O ******************************************************************/
  void y_or    (long,long); // ou                                          97
  void ori     (long);      //                                             98
  void orm     (long);      //                                             99
  /**** R ******************************************************************/
  void rotcl   (long);      // rotation gauche                            100
  void rotcr   (long);      // rotation droite                            101
  void rotl    (long);      // rotation gauche                            102
  void rotr    (long);      // rotation droite                            103
  void rte     (void);      //                                            104
  void rts     (void);      //                                            105
  /**** S ******************************************************************/
  void sett    (void);      // T <- 1                                     106
  void shal    (long);      // dcalage gauche                            107
  void shar    (long);      // dcalage droite                            108
  void shll    (long);      // dcalage gauche                            109
  void shll2   (long);      // dcalage gauche                            110
  void shll8   (long);      // dcalage gauche                            111
  void shll16  (long);      // dcalage gauche                            112
  void shlr    (long);      // dcalage droite                            113
  void shlr2   (long);      // dcalage droite                            114
  void shlr8   (long);      // dcalage droite                            115
  void shlr16  (long);      // dcalage droite                            116
  void sleep   (void);      // fait dodo...                               117
  void stcgbr  (long);      //                                            118
  void stcmgbr (long);      //                                            119
  void stcmsr  (long);      //                                            120
  void stcmvbr (long);      //                                            121
  void stcsr   (long);      //                                            122
  void stcvbr  (long);      //                                            124
  void stsmach (long);      //                                            125
  void stsmacl (long);      //                                            126
  void stsmmach(long);      //                                            127
  void stsmmacl(long);      //                                            128
  void stsmpr  (long);      //                                            129
  void stspr   (long);      //                                            130
  void sub     (long,long); // soustraction                               131
  void subc    (long,long); // soustraction avec retenue                  132
  void subv    (long,long); // soustraction avec vrification             133
  void swapb   (long,long); //                                            134
  void swapw   (long,long); //                                            135
  /**** T ******************************************************************/
  void tas     (long);      //                                            136
  void trapa   (long);      //                                            137
  void tst     (long,long); //                                            138
  void tsti    (long);      //                                            139
  void tstm    (long);      //                                            140
  /**** X ******************************************************************/
  void y_xor   (long,long); // le shrif de l'espace                      141
  void xori    (long);      //                                            142
  void xorm    (long);      //                                            143
  void xtrct   (long,long); //                                            144
};

#endif
