/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "registres.hh"
#include "sem.hh"
#include "message.hh"
#include "superh.hh"

int Memoire::v = 0;
FileMessage Memoire::fm(SATURNE_MSG);

Memoire::Memoire(unsigned long taille, char cle) {
  this->taille = taille;
  if (taille == 0) return;
  int clef = ftok("/usr/",numero = cle);
  int shmid = shmget(clef,taille,IPC_CREAT | 0600);
  memoire = (unsigned char *) shmat(shmid,NULL,0);

  Message message(cle);
  fm.envoyer(message);
}

Memoire::~Memoire(void) {
  if (taille == 0) return;
  int clef = ftok("/usr/",numero);
  int shmid = shmget(clef,taille,IPC_CREAT | 0600);
  shmdt(memoire);

  Message message(numero);
  fm.recevoir(message, numero);
  if (!fm.recevoir(message, numero)) shmctl(shmid,IPC_RMID,NULL);
  else fm.envoyer(message);
}

unsigned char Memoire::getOctet(unsigned long adr) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  return memoire[adr];
}

void Memoire::setOctet(unsigned long adr, unsigned char valeur) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  memoire[adr] = valeur;
}

unsigned short Memoire::getMot(unsigned long adr) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  return (unsigned short) ((memoire[adr] << 8) | memoire[adr + 1]);
}

void Memoire::setMot(unsigned long adr, unsigned short valeur) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  memoire[adr] = (unsigned char) (valeur >> 8);
  memoire[adr + 1] = (unsigned char) (valeur & 0x00FF);
}

unsigned long Memoire::getGrosmot(unsigned long adr) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  return (unsigned long) ((memoire[adr] << 24) |
			  (memoire[adr + 1] << 16) |
			  (memoire[adr + 2] << 8) |
			  memoire[adr + 3]);
}

void Memoire::setGrosmot(unsigned long adr, unsigned long valeur) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  memoire[adr] = (unsigned char) (valeur >> 24);
  memoire[adr + 1] = (unsigned char) ((valeur >> 16) & 0x00FF);
  memoire[adr + 2] = (unsigned char) ((valeur >> 8) & 0x00FF);
  memoire[adr + 3] = (unsigned char) (valeur & 0x00FF);
}

void Memoire::charger(char *fichier, unsigned long adr) {
  struct stat infos;

  if (stat(fichier,&infos) == -1)
    throw FichierInexistant(fichier);

  if (((unsigned int) infos.st_size) > taille) {
    cerr << "\nFichier " << fichier << " trop grand pour tre charg\n";
    throw MauvaisAcces(adr); // FIXME ce n'est pas vraiment un mauvais accs
  }

  ifstream exe(fichier, ios::in | ios::binary);
  exe.read(((char * )memoire) + adr, infos.st_size);
  exe.close();
}

ostream& operator<<(ostream& os, const Memoire& mem) {
  for(unsigned long i = 0;i < mem.taille;i++) {
    os << hex << setw(6) << i << " -> " << setw(10) << mem.memoire[i] << endl;
  }
  return os;
}

Registre::Registre(const char *n, unsigned long t, char c) : Memoire(t, c) {
  strcpy(nom, n);
}

unsigned char Registre::getOctet(unsigned long adr) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  cerr << nom << "\t: " << setw(10) << adr
	      << " <- " << setw(10) << (int) memoire[adr] << '\n';
  return memoire[adr];
}

void Registre::setOctet(unsigned long adr, unsigned char valeur) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  cerr << nom << "\t: " << setw(10) << adr
	      << " <- " << setw(10) << (int) valeur << '\n';
  memoire[adr] = valeur;
}

unsigned short Registre::getMot(unsigned long adr) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  unsigned short retour = ((memoire[adr] << 8) | memoire[adr + 1]);
  cerr << nom << "\t: " << setw(10) << adr
	      << " <- " << setw(10) << retour << '\n';
  return retour;
}

void Registre::setMot(unsigned long adr, unsigned short valeur) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  cerr << nom << "\t: " << setw(10) << adr
	      << " <- " << setw(10) << valeur << '\n';
  memoire[adr] = (unsigned char) (valeur >> 8);
  memoire[adr + 1] = (unsigned char) (valeur & 0x00FF);
}

unsigned long Registre::getGrosmot(unsigned long adr) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  unsigned long retour = ((memoire[adr] << 24) |
			  (memoire[adr + 1] << 16) |
			  (memoire[adr + 2] << 8) |
			  memoire[adr + 3]);
  cerr << nom << "\t: " << setw(10) << adr
	      << " <- " << setw(10) << retour << '\n';
  return retour;
}

void Registre::setGrosmot(unsigned long adr, unsigned long valeur) {
  if (adr >= taille) { throw MauvaisAcces(adr); }
  cerr << nom << "\t: " << setw(10) << adr
	      << " <- " << setw(10) << valeur << '\n';
  memoire[adr] = (unsigned char) (valeur >> 24);
  memoire[adr + 1] = (unsigned char) ((valeur >> 16) & 0x00FF);
  memoire[adr + 2] = (unsigned char) ((valeur >> 8) & 0x00FF);
  memoire[adr + 3] = (unsigned char) (valeur & 0x00FF);
}

/*********************/
/*                   */
/* Mmoire Saturn    */
/*                   */
/*********************/


MemoireSaturn::MemoireSaturn(void) : Memoire(0,MEMSAT) {
  rom         = new Memoire(0x80000, ROM);
  smpc        = new RegistresSmpc();
  ram         = new Memoire(0x10000, RAM);
  ramLow      = new Memoire(0x100000, RAMLOW);
  cs1         = new Cs1();
  cs2         = new Cs2();
  sound       = new Memoire(0x7FFFF, SOUND);
  soundr      = new Memoire(0xEE4, SOUNDR);
  vdp1_1      = new Registre("memvdp1", 0xC0000, VDP1_1);
  vdp1_2      = new RegistresVdp1();
  vdp2_1      = new Memoire(0x80000, VDP2_1);
  vdp2_2      = new Memoire(0x1000, VDP2_2);
  vdp2_3      = new RegistresVdp2();
  scu         = new RegistresScu();
  ramHigh     = new Memoire(0x100000, RAMHIGH);
  purgeArea   = new Dummy();
  adressArray = new Memoire(0x3FF, ADRESSARRAY);
  modeSdram   = new Memoire(0x4FFF, MODESDRAM);
  onchip      = new Memoire(0x1FF, ONCHIP);
}

MemoireSaturn::~MemoireSaturn(void) {
  delete rom;
  delete smpc;
  delete ram;
  delete ramLow;
  delete cs1;
  delete cs2;
  delete sound;
  delete soundr;
  delete vdp1_1;
  delete vdp1_2;
  delete vdp2_1;
  delete vdp2_2;
  delete vdp2_3;
  delete scu;
  delete ramHigh;
  delete purgeArea;
  delete adressArray;
  delete modeSdram;
  delete onchip;
}

void MemoireSaturn::mappage(unsigned long adr) {
  switch (adr >> 29) {
  case 0:
  case 1: {
    unsigned long nadr = adr & 0x1FFFFFFF;
    if (nadr < 0x80000) { mapMem = rom; mapAdr = nadr; return; }
    if ((nadr >=  0x100000) && (nadr <  0x100080)) { mapMem = smpc;    mapAdr = nadr & 0x000000FF; return;}
    if ((nadr >=  0x180000) && (nadr <  0x190000)) { mapMem = ram;     mapAdr = nadr & 0x0000FFFF; return;}
    if ((nadr >=  0x200000) && (nadr <  0x300000)) { mapMem = ramLow;  mapAdr = nadr & 0x000FFFFF; return;}
    if ((nadr >= 0x4000000) && (nadr < 0x5000000)) { mapMem = cs1;     mapAdr = nadr & 0x00FFFFFF; return;}
    if ((nadr >= 0x5800000) && (nadr < 0x5900000)) { mapMem = cs2;     mapAdr = nadr & 0x000FFFFF; return;}
    if ((nadr >= 0x5A00000) && (nadr < 0x5A80000)) { mapMem = sound;   mapAdr = nadr & 0x0000FFFF; return;}
    if ((nadr >= 0x5B00000) && (nadr < 0x5B00EE4)) { mapMem = soundr;  mapAdr = nadr & 0x00000FFF; return;}
    if ((nadr >= 0x5C00000) && (nadr < 0x5CC0000)) { mapMem = vdp1_1;  mapAdr = nadr & 0x000FFFFF; return;}
    if ((nadr >= 0x5D00000) && (nadr < 0x5D00018)) { mapMem = vdp1_2;  mapAdr = nadr & 0x000000FF; return;}
    if ((nadr >= 0x5E00000) && (nadr < 0x5E80000)) { mapMem = vdp2_1;  mapAdr = nadr & 0x000FFFFF; return;}
    if ((nadr >= 0x5F00000) && (nadr < 0x5F01000)) { mapMem = vdp2_2;  mapAdr = nadr & 0x0000FFFF; return;}
    if ((nadr >= 0x5F80000) && (nadr < 0x5F80120)) { mapMem = vdp2_3;  mapAdr = nadr & 0x00000FFF; return;}
    if ((nadr >= 0x5FE0000) && (nadr < 0x5FE00D0)) { mapMem = scu;     mapAdr = nadr & 0x000000FF; return;}
    if ((nadr >= 0x6000000) && (nadr < 0x6100000)) { mapMem = ramHigh; mapAdr = nadr & 0x000FFFFF; return;}
    }
    throw MauvaisAcces(adr);
  case 2:
    mapMem = purgeArea; mapAdr = adr; return;
  case 3: {
    unsigned long nadr = adr & 0x1FFFFFFF;
    if (nadr < 0x3FF) { mapMem = adressArray; mapAdr = nadr; return;}
    throw MauvaisAcces(adr);
  }
  case 7:
    if ((adr >= 0xFFFF8000) && (adr < 0xFFFFC000)) {mapMem = modeSdram; mapAdr = adr & 0x00000FFF; return;}
    if (adr >= 0xFFFFFE00) { mapMem = onchip; mapAdr = adr & 0x000001FF; return;}
    throw MauvaisAcces(adr);
  default:
    throw MauvaisAcces(adr);
  }
}

unsigned char MemoireSaturn::getOctet(unsigned long adr) {
  mappage(adr);
  return mapMem->getOctet(mapAdr);
}

void MemoireSaturn::setOctet(unsigned long adr, unsigned char valeur) {
  mappage(adr);
  mapMem->setOctet(mapAdr, valeur);
}

unsigned short MemoireSaturn::getMot(unsigned long adr) {
  mappage(adr);
  return mapMem->getMot(mapAdr);
}

void MemoireSaturn::setMot(unsigned long adr, unsigned short valeur) {
  mappage(adr);
  mapMem->setMot(mapAdr, valeur);
}

unsigned long MemoireSaturn::getGrosmot(unsigned long adr) {
  mappage(adr);
  return mapMem->getGrosmot(mapAdr);
}

void MemoireSaturn::setGrosmot(unsigned long adr, unsigned long valeur) {
  mappage(adr);
  mapMem->setGrosmot(mapAdr, valeur);
}

void MemoireSaturn::chargerBios(char *fichier) {
  rom->charger(fichier, 0);
}

void MemoireSaturn::charger(char *fichier, unsigned long adr) {
  mappage(adr);
  mapMem->charger(fichier, mapAdr);
}
