/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MEMOIRE_HH
#define MEMOIRE_HH

#include "exception.hh"
#include "objet.hh"

class Memoire {
private:
protected:
  char numero;
  unsigned long taille;
  unsigned char *memoire;
  static int v;
  static FileMessage fm;
public:
  Memoire(unsigned long, char);
  virtual ~Memoire(void);

  virtual unsigned char  getOctet   (unsigned long);
  virtual void           setOctet   (unsigned long, unsigned char);
  virtual unsigned short getMot     (unsigned long);
  virtual void           setMot     (unsigned long, unsigned short);
  virtual unsigned long  getGrosmot (unsigned long);
  virtual void           setGrosmot (unsigned long, unsigned long);

  virtual void charger(char *, unsigned long);

  friend ostream& operator<<(ostream&, const Memoire&);
};

class Registre : public Memoire {
protected:
  char nom[11];
public:
  Registre(const char *, unsigned long, char);
  unsigned char  getOctet   (unsigned long);
  void           setOctet   (unsigned long, unsigned char);
  unsigned short getMot     (unsigned long);
  void           setMot     (unsigned long, unsigned short);
  unsigned long  getGrosmot (unsigned long);
  void           setGrosmot (unsigned long, unsigned long);
};

class Dummy : public Memoire {
public:
  Dummy(void) : Memoire(0, DUMMY) {}
  ~Dummy(void) {}
  unsigned char getOctet(unsigned long) { return 0; }
  void setOctet(unsigned long, unsigned char) {}
  unsigned short getMot(unsigned long) { return 0; }
  void setMot(unsigned long, unsigned short) {}
  unsigned long getGrosmot(unsigned long) { return 0; }
  void setGrosmot(unsigned long, unsigned long) {}
};

class MemoireSaturn : public Memoire {
private:
  Memoire *rom;		//        0 -    80000
  Memoire *smpc;	//   100000 -   100080
  Memoire *ram;		//   180000 -   190000
  Memoire *ramLow;      //   200000 -   300000
  Memoire *minit;       //  1000000 -  1000004
  Memoire *sinit;       //  1800000 -  1800004
  Memoire *cs0;         //  2000000 -  4000000
  Memoire *cs1;         //  4000000 -  5000000
  Memoire *dummy;       //  5000000 -  5800000
  Memoire *cs2;         //  5800000 -  5900000
  Memoire *sound;       //  5A00000 -  5A80000
  Memoire *soundr;	//  5B00000 -  5B00EE4
  Memoire *vdp1_1;      //  5C00000 -  5CC0000
  Memoire *vdp1_2;	//  5D00000 -  5D00018
  Memoire *vdp2_1;      //  5E00000 -  5E80000
  Memoire *vdp2_2;      //  5F00000 -  5F01000
  Memoire *vdp2_3;	//  5F80000 -  5F80120
  Memoire *scu;		//  5FE0000 -  5FE00D0
  Memoire *ramHigh;     //  6000000 -  6100000
  Memoire *purgeArea;	// 40000000 - 50000000
  Memoire *adressArray; // 60000000 - 600003FF
  Memoire *modeSdram;   // FFFF8000 - FFFFBFFF
  Memoire *onchip;	// FFFFFE00 - FFFFFFFF

  Memoire *mapMem;
  unsigned long mapAdr;
  void mappage(unsigned long);

public:
  MemoireSaturn(void);
  ~MemoireSaturn(void);

  unsigned char  getOctet   (unsigned long);
  void           setOctet   (unsigned long, unsigned char);
  unsigned short getMot     (unsigned long);
  void           setMot     (unsigned long, unsigned short);
  unsigned long  getGrosmot (unsigned long);
  void           setGrosmot (unsigned long, unsigned long);

  void           charger    (char *,unsigned long);

  void chargerBios(char *);
};

#endif
