/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef REVEIL_HH
#define REVEIL_HH

#include <sys/types.h>
#include "saturn.hh"
#include "sem.hh"

template<class U> class Reveil {
private:
  static U *u;
public:
  static void init(void);
  static pid_t& pid(void);
  static void lancer(void);
  static void gererSignal(int);
};

template<class U> U *Reveil<U>::u;

template<class U> void Reveil<U>::init(void) {
  u = new U();
}

template<class U> pid_t& Reveil<U>::pid(void) {
  return u->pid();
}

template<class U> void Reveil<U>::lancer(void) {
  Semaphore sem(PROC_SEM);
  sigset_t masque1;
  sigfillset(&masque1);
  sigprocmask(SIG_BLOCK, &masque1, NULL);

  struct sigaction action;
  action.sa_handler = &gererSignal;
  action.sa_mask = masque1;

  sigaction(SIGUSR1, &action, NULL);
  sigaction(SIGUSR2, &action, NULL);
  sigaction(SIGALRM, &action, NULL);

  sem.V();

  sigset_t masque;
  sigemptyset(&masque);
  sigsuspend(&masque);
}

template<class U> void Reveil<U>::gererSignal(int signum) {
  sigset_t attente, masque;
  sigaction(SIGUSR1, NULL, NULL);
  sigaction(SIGUSR2, NULL, NULL);
  sigaction(SIGALRM, NULL, NULL);
  switch(signum) {
  case SIGUSR1:
    u->executer();
    break;
  case SIGUSR2:
    delete u;
    return;
  case SIGALRM:
    u->appelPeriodique();
    alarm(1);
    break;
  }
  sigpending(&attente);
  sigemptyset(&masque);
  sigsuspend(&masque);
}

#endif
