/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "superh.hh"
#include "smpc.hh"
#include "cs2.hh"
#include "reveil.hh"
#include "intc.hh"
#include "vdp1.hh"
#include "vdp2.hh"

class Yabause {
private:
  static Semaphore sem;
public:
  Yabause(void) {
    sem.init(0);
  }
  ~Yabause(void) {
    sem.supprimer();
  }
  template<class U> pid_t demarrer(void) {
    Reveil<U>::init();
    switch (Reveil<U>::pid() = fork()) {
    case 0:
      Reveil<U>::lancer();
      exit(0);
    }
    sem.P();
    return Reveil<U>::pid();
  }
  template<class U> void stopper(void) {
    kill(Reveil<U>::pid(), SIGUSR2);
  }
};

Semaphore Yabause::sem(PROC_SEM);

int main(int argc, char **argv) {
  // test des arguments, trs primaire pour l'instant...
  if (argc < 2) {
    cerr << "usage: " << argv[0] << " <bios>\n";
    return 1;
  }

  cerr << hex;

  Yabause yabause;

  yabause.demarrer<CD>();
  yabause.demarrer<Vdp1>();
  yabause.demarrer<Vdp2>();
  yabause.demarrer<Smpc>(); /* Smpc must be launch after others processors
			       to "know" them */

  switch (fork()) {
    case 0: {
      Intc intc;
      intc.lancer();
      exit(0);
    }
  }

  MemoireSaturn mem;

  // on charge le bios
  try { mem.chargerBios(argv[1]); }
  catch (FichierInexistant fi) {
    cerr << '\n' << fi << '\n';
    return 1;
  }

  // maintenant on a un bios on peut crer le processeur
  MasterSH *proc = new MasterSH(mem);

  // et lancer l'excution
  try {
    proc->lancer();
  }
  catch (InstructionIllegale il) {
    cerr << "Emulation stoppe sur le code: " <<  il << '\n';
  }
  catch (InstructionNonImplementee ini) { // FIXME: devrait disparaitre
    cerr << '\n' << ini.nom << " pas implemente\n";
  }
  catch (MauvaisAcces ma) {
    cerr << '\n' << ma << '\n';
  }

  cerr << *proc;

  delete proc;
  yabause.stopper<Vdp2>();
  yabause.stopper<Vdp1>();
  yabause.stopper<Smpc>();
  yabause.stopper<CD>();
  //yabause.stopper<Intc>();
  FileMessage fm(INTC_MSG);
  Message message(Intc::INTC_STOP);
  fm.envoyer(message);
}
