/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "smpc.hh"
#include "cs2.hh"

RegistresSmpc::RegistresSmpc(void) : Registre("smpc", 0x80, SMPC) {}

unsigned char RegistresSmpc::getIREG(int i) {return memoire[1 + 2 * i];}
unsigned char RegistresSmpc::getCOMREG(void) {return memoire[0x1F];}
unsigned char RegistresSmpc::getOREG(int i) {return memoire[0x21 + 2 * i];}
unsigned char RegistresSmpc::getSR(void) {return memoire[0x61];}
unsigned char RegistresSmpc::getSF(void) {return memoire[0x63];}
void RegistresSmpc::setIREG(int i, unsigned char val) {memoire[1+2*i] = val;}
void RegistresSmpc::setCOMREG(unsigned char val) {memoire[0x1F] = val;}
void RegistresSmpc::setOREG(int i, unsigned char val) {memoire[0x21+2*i]= val;}
void RegistresSmpc::setSR(unsigned char val) {memoire[0x61] = val;}
void RegistresSmpc::setSF(unsigned char val) {memoire[0x63] = val;}  

void RegistresSmpc::setOctet(unsigned long adr, unsigned char valeur) {
  cerr << hex << "smpc\t: " << setw(10) << adr << " <- "
                            << setw(10) << (int) valeur << endl;
  if (adr >= taille) { throw MauvaisAcces(adr); }
  memoire[adr] = valeur;

  if (adr == 0x1F) {
    kill(Reveil<Smpc>::pid(), SIGUSR1);
  }
}

Smpc::Smpc(void) {
  dotsel = false;
  mshnmi = false;
  sndres = false;
  cdres = false;
  sysres = false;
  resb = false;
  ste = false;
  resd = false;
}

void Smpc::executer(void) {
  switch(memoire.getCOMREG()) {
  case 0x6:
    cerr << "smpc\t: CDON\n";
    CDON();
    break;
  case 0x7:
    cerr << "smpc\t: CDOFF\n";
    CDOFF();
    break;
  case 0x10:
    cerr << "smpc\t: INTBACK\n";
    INTBACK();
    break;
  case 0x19:
    cerr << "smpc\t: RESENAB\n";
    RESENAB();
    break;
  case 0x1A:
    cerr << "smpc\t: RESDISA\n";
    RESDISA();
    break;
  default:
    cerr << "smpc\t: " << ((int) memoire.getCOMREG())
	 << " commande non implante\n";
  }
  
  memoire.setSF(0);
}

void Smpc::INTBACK(void) {
  //if (memoire->getIREG(0)) {
    // renvoyer les donnes d'heure, code cartouche, code de zone, ...
    // Ecrire les donnes dans SR

    bool resteDonnees = false;
    memoire.setSR(0x40 | (resteDonnees << 5) | (resb << 5));

    // Ecrire les donnes dans OREG0
    
    // Ecrire les donnes dans OREG1-7

    time_t tmp = time(NULL);
    struct tm *temps = localtime(&tmp);
    unsigned char annee[4] = {
      (1900 + temps->tm_year) / 1000,
      ((1900 + temps->tm_year) % 1000) / 100,
      (((1900 + temps->tm_year) % 1000) % 100) / 10,
      (((1900 + temps->tm_year) % 1000) % 100) % 10 };
    memoire.setOREG(1, (annee[0] << 4) | annee[1]);
    memoire.setOREG(2, (annee[2] << 4) | annee[3]);
    memoire.setOREG(3, (temps->tm_wday << 4) | (temps->tm_mon + 1));
    memoire.setOREG(4, ((temps->tm_mday / 10) << 4) | (temps->tm_mday % 10));
    memoire.setOREG(5, ((temps->tm_hour / 10) << 4) | (temps->tm_hour % 10));
    memoire.setOREG(6, ((temps->tm_min / 10) << 4) | (temps->tm_min % 10));
    memoire.setOREG(7, ((temps->tm_sec / 10) << 4) | (temps->tm_sec % 10));
    
    // Ecrire le code cartouche dans OREG8, bit 0 et 1
    
    memoire.setOREG(8, 0); // FIXME : du vrai hasard :)
    
    // Ecrire le code de zone dans OREG9, bit 0-7
    // 1 -> japon
    // 2 -> asie ntsc
    // 4 -> amrique du nord
    // 5 -> amrique du sud/centrale ntsc
    // 6 -> core
    // A -> asie pal
    // C -> europe + divers pal
    // D -> amrique du sud/centrale pal
    
    memoire.setOREG(9, 0xC);
    
    // Etat du systme, premire partie dans OREG10, bit 0-7
    // bit | valeur | commentaires
    // ---------------------------
    // 7   | 0      |
    // 6   | DOTSEL |
    // 5   | 1      |
    // 4   | 1      |
    // 3   | MSHNMI |
    // 2   | 1      |
    // 1   | SYSRES | 
    // 0   | SNDRES |
    
    // FIXME
    memoire.setOREG(10, 0x34|(dotsel<<6)|(mshnmi<<3)|(sysres<<1)|sndres);
    
    // Etat du systme, deuxime partie dans OREG11, bit 6
    // bit 6 -> CDRES
    
    memoire.setOREG(11, cdres << 6); // FIXME
    
    // Les sauvegardes vont dans OREG12-15

    for(int i = 0;i < 4;i++) memoire.setOREG(12, 0);
    
    // Ecrire les donnes dans OREG31

    memoire.setOREG(31, 0);
    
  /*}
  else {
    // ne pas renvoyer les donnes d'heure, code cartouche, code de zone, ...
  }*/
}

void Smpc::RESENAB(void) {
  resd = false;
  memoire.setOREG(31, 0x19);
}

void Smpc::RESDISA(void) {
  resd = true;
  memoire.setOREG(31, 0x1A);
}

void Smpc::CDON(void) {
}

void Smpc::CDOFF(void) {
}
