/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "superh.hh"
#include "message.hh"
#include "sem.hh"
#include "vdp1.hh"
#include "vdp2.hh"

SuperH::SuperH(void) {
}

SuperH::SuperH(Memoire& mem) {
  memoire = &mem;
  SR.partie.T = SR.partie.S = SR.partie.Q = SR.partie.M = 0;
  SR.partie.I = 0xF;
  SR.partie.inutile1 = 0;
  SR.partie.inutile2 = 0;
  VBR = 0;

  PC = memoire->getGrosmot(0) + 4;

  R[15] = memoire->getGrosmot(4);

  stop = false;
  interruption = false;
  _delai = 0;
  for(unsigned short i = 0;i < 0xFFFF;i++) {
    instruction = i;
    opcodes[i] = decode();
  }
}

void SuperH::lancer(void) {
  unsigned long i = 0; // FIXME nom de variable trop explicite ;)
  while (!stop) {
    if (_delai) {
      unsigned long tmp = PC;
      PC = _delai;
      executer();
      PC = tmp;
      _delai = 0;
    }
    else {
      if (interruption) {
	interruption = false;
        R[15] -= 4;
	memoire->setGrosmot(R[15], SR.tout);
	R[15] -= 4;
	memoire->setGrosmot(R[15], PC);
	SR.partie.I = niveau;
	PC = memoire->getGrosmot(VBR + (vecteur << 2)) + 4;
      }
      executer();
    }
    i++; // FIXME compte les instructions pas les cycles
    switch(i) {
    case 223951:
      kill(Reveil<Vdp2>::pid(), SIGUSR1); // tue yabause, pourquoi ?
      break;
    case 447902:
      kill(Reveil<Vdp2>::pid(), SIGUSR1); // tue yabause, pourquoi ?
      kill(Reveil<Vdp1>::pid(), SIGUSR1);
      i = 0;
      break;
    }
  }
}

void SuperH::executer(void) {
  instruction = memoire->getMot(PC - 4);

  (this->*opcodes[instruction])();
}

ostream& operator<<(ostream& os, const SuperH& sh) {
  for(int j = 0;j < 4;j++) {
    os << "\t";
    for(int i = 4*j;i < 4 * (j + 1);i++) {
      os << "R" << dec << left << setw(4) << i
	        << hex << right << setw(10) << sh.R[i] << "  ";
    }
    os << "\n";
  }

  os << dec << left << "\tSR    M  " << setw(11) << sh.SR.partie.M
                            << "Q  " << setw(11) << sh.SR.partie.Q
                     << hex << "I  " << setw(11) << sh.SR.partie.I
                     << dec << "S  " << setw(11) << sh.SR.partie.S
                            << "T  " <<             sh.SR.partie.T << endl;

  os << hex << right << "\tGBR\t" << setw(10) << sh.GBR
                     << "\tMACH\t" << setw(10) << sh.MACH
                     << "\tPR\t" << setw(10) << sh.PR << endl
                     << "\tVBR\t" << setw(10) << sh.VBR
                     << "\tMACL\t" << setw(10) << sh.MACL
                     << "\tPC\t" << setw(10) << sh.PC << endl;

  return os;
}

void SuperH::undecoded(void) {
  throw InstructionIllegale(instruction);
}

void SuperH::add(void) {
  R[Instruction::b(instruction)] += R[Instruction::c(instruction)];
  PC += 2;
}

void SuperH::addi(void) {
  long source = Instruction::cd(instruction);
  long dest = Instruction::b(instruction);

  if (source & 128)
    R[dest] += (0xFFFFFF00 | (long) source);
  else
    R[dest] += (0x000000FF & (long) source);
  PC += 2;
}

void SuperH::addc(void) {
  unsigned long tmp0, tmp1;
  long source = Instruction::c(instruction);
  long dest = Instruction::b(instruction);

  tmp1 = R[source] + R[dest];
  tmp0 = R[dest];

  R[dest] = tmp1 + SR.partie.T;
  if (tmp0 > tmp1)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  if (tmp1 > R[dest])
    SR.partie.T = 1;
  PC += 2;
}

void SuperH::addv(void) {
  long i,j,k;
  long source = Instruction::c(instruction);
  long dest = Instruction::b(instruction);

  if ((long) R[dest] >= 0) i = 0; else i = 1;
  //i = ((long) R[dest] < 0);
  
  if ((long) R[source] >= 0) j = 0; else j = 1;
  //j = ((long) R[source] < 0);
  
  j += i;
  R[dest] += R[source];

  if ((long) R[dest] >= 0) k = 0; else k = 1;
  //k = ((long) R[dest] < 0);

  k += j;
  
  if (j == 0 || j == 2)
    if (k == 1) SR.partie.T = 1;
    else SR.partie.T = 0;
  else SR.partie.T = 0;
  //SR.partie.T = ((j == 0 || j == 2) && (k == 1));
  PC += 2;
}

void SuperH::y_and(void) {
  R[Instruction::b(instruction)] &= R[Instruction::c(instruction)];
  PC += 2;
}

void SuperH::andi(void) {
  R[0] &= (0x000000FF & (long) Instruction::cd(instruction));
  PC += 2;
}

void SuperH::andm(void) {
  long temp;
  long source = Instruction::cd(instruction);

  temp = (long) memoire->getOctet(GBR + R[0]);
  temp &= (0x000000FF & (long) source);
  memoire->setOctet((GBR + R[0]),temp);
  PC += 2;
}
 
void SuperH::bf(void) { // FIXME peut etre amelior
  long disp;
  long d = Instruction::cd(instruction);

  if ((d&0x80)==0)
    disp=(0x000000FF & (long)d);
  else
    disp=(0xFFFFFF00 | (long)d);
  if (SR.partie.T == 0)
    PC = PC+(disp<<1)+4;
  else PC+=2;
}

void SuperH::bfs(void) { // FIXME peut tre amlior
  long disp;
  unsigned long temp;
  long d = Instruction::cd(instruction);

  temp = PC;
  if ((d & 0x80) == 0)
    disp = (0x000000FF & (long) d);
  else
    disp = (0xFFFFFF00 | (long) d);
  if (SR.partie.T == 0) {
    PC = PC + (disp << 1) + 4;

    _delai = temp + 2;
  }
  else {
    PC += 2;

  }
}

void SuperH::bra(void) {
  unsigned long temp;
  long d = Instruction::bcd(instruction);

  long disp;
  if ((d&0x800)==0) disp=(0x00000FFF & d);
  else disp=(0xFFFFF000 | d);
  temp=PC;
  PC = PC + (disp<<1) + 4;

  _delai = temp + 2;
}

void SuperH::braf(void) {
  unsigned long temp;
  long m = Instruction::b(instruction);

  temp = PC;
  PC += R[m] + 4; /* FIXME : Correction trouve dans semu, est-ce bon ?
		     Hyperion est d'accord ! */

  _delai = temp + 2;
}

void SuperH::bsr(void) {
  long disp;
  long d = Instruction::bcd(instruction);

  if ((d&0x800)==0) disp = (0x00000FFF & d);
  else disp = (0xFFFFF000 | d);
  PR = PC;
  PC = PC+(disp<<1) + 4;

  _delai = PR + 2;
}

void SuperH::bsrf(void) {
  PR = PC;
  PC += R[Instruction::b(instruction)] + 4; // correction donne par Hyperion
  _delai = PR + 2;
}

void SuperH::bt(void) { // FIXME ya plus rapide
  long disp;
  long d = Instruction::cd(instruction);

  if ((d&0x80)==0)
    disp=(0x000000FF & (long)d);
  else
    disp=(0xFFFFFF00 | (long)d);
  if (SR.partie.T == 1)
    PC = PC+(disp<<1)+4;
  else
    PC += 2;
}

void SuperH::bts(void) {
  long disp;
  unsigned long temp;
  long d = Instruction::cd(instruction);
  
  if (SR.partie.T) {
    temp = PC;
    if ((d & 0x80) == 0) disp = (0x000000FF & (long) d);
    else disp = (0xFFFFFF00 | (long)d);
    PC += (disp << 1) + 4;
    _delai = temp + 2;
  }
  else PC+=2;
}

void SuperH::clrmac(void) {
  MACH = 0;
  MACL = 0;
  PC += 2;
}

void SuperH::clrt(void) {
  SR.partie.T = 0;
  PC += 2;
}

void SuperH::cmpeq(void) {
  if (R[Instruction::b(instruction)] == R[Instruction::c(instruction)])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC += 2;
}

void SuperH::cmpge(void) {
  if ((long)R[Instruction::b(instruction)] >=
		  (long)R[Instruction::c(instruction)])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC += 2;
}

void SuperH::cmpgt(void) {
  if ((long)R[Instruction::b(instruction)]>(long)R[Instruction::c(instruction)])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmphi(void) {
  if ((unsigned long)R[Instruction::b(instruction)]>
		  (unsigned long)R[Instruction::c(instruction)])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmphs(void) {
  if ((unsigned long)R[Instruction::b(instruction)]>=
		  (unsigned long)R[Instruction::c(instruction)])
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmpim(void) {
  long imm;
  long i = Instruction::cd(instruction);

  if ((i&0x80)==0)
    imm = (0x000000FF & (long) i);
  else
    imm=(0xFFFFFF00 | (long) i);
  if (R[0] == (unsigned long) imm) // FIXME: ouais a doit tre bon...
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmppl(void) {
  if ((long)R[Instruction::b(instruction)]>0)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmppz(void) {
  if ((long)R[Instruction::b(instruction)]>=0)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::cmpstr(void) {
  unsigned long temp;
  long HH,HL,LH,LL;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  
  temp=R[n]^R[m];
  HH = (temp>>12) & 0x000000FF;
  HL = (temp>>8) & 0x000000FF;
  LH = (temp>>4) & 0x000000FF;
  LL = temp & 0x000000FF;
  HH = HH && HL && LH && LL;
  if (HH == 0)
    SR.partie.T = 1;
  else
    SR.partie.T = 0;
  PC+=2;
}

void SuperH::div0s(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  if ((R[n]&0x80000000)==0)
    SR.partie.Q = 0;
  else
    SR.partie.Q = 1;
  if ((R[m]&0x80000000)==0)
    SR.partie.M = 0;
  else
    SR.partie.M = 1;
  SR.partie.T = !(SR.partie.M == SR.partie.Q);
  PC += 2;
}

void SuperH::div0u(void) {
  SR.partie.M = SR.partie.Q = SR.partie.T = 0;
  PC += 2;
}

void SuperH::div1(void) {
  unsigned long tmp0;
  unsigned char old_q, tmp1;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  
  old_q = SR.partie.Q;
  SR.partie.Q = (unsigned char)((0x80000000 & R[n])!=0);
  R[n] <<= 1;
  R[n]|=(unsigned long) SR.partie.T;
  switch(old_q){
  case 0:
    switch(SR.partie.M){
    case 0:
      tmp0 = R[n];
      R[n] -= R[m];
      tmp1 = (R[n] > tmp0);
      switch(SR.partie.Q){
      case 0:
	SR.partie.Q = tmp1;
	break;
      case 1:
	SR.partie.Q = (unsigned char) (tmp1 == 0);
	break;
      }
      break;
    case 1:
      tmp0 = R[n];
      R[n] += R[m];
      tmp1 = (R[n] < tmp0);
      switch(SR.partie.Q){
      case 0:
	SR.partie.Q = (unsigned char) (tmp1 == 0);
	break;
      case 1:
	SR.partie.Q = tmp1;
	break;
      }
      break;
    }
    break;
  case 1:switch(SR.partie.M){
  case 0:
    tmp0 = R[n];
    R[n] += R[m];
    tmp1 = (R[n] < tmp0);
    switch(SR.partie.Q){
    case 0:
      SR.partie.Q = tmp1;
      break;
    case 1:
      SR.partie.Q = (unsigned char) (tmp1 == 0);
      break;
    }
    break;
  case 1:
    tmp0 = R[n];
    R[n] -= R[m];
    tmp1 = (R[n] > tmp0);
    switch(SR.partie.Q){
    case 0:
      SR.partie.Q = (unsigned char) (tmp1 == 0);
      break;
    case 1:
      SR.partie.Q = tmp1;
      break;
    }
    break;
  }
  break;
  }
  SR.partie.T = (SR.partie.Q == SR.partie.M);
  PC += 2;
}


void SuperH::dmuls(void) {
  unsigned long RnL,RnH,RmL,RmH,Res0,Res1,Res2;
  unsigned long temp0,temp1,temp2,temp3;
  long tempm,tempn,fnLmL;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  
  tempn = (long)R[n];
  tempm = (long)R[m];
  if (tempn < 0) tempn = 0 - tempn;
  if (tempm < 0) tempm = 0 - tempm;
  if ((long) (R[n] ^ R[m]) < 0) fnLmL = -1;
  else fnLmL = 0;
  
  temp1 = (unsigned long) tempn;
  temp2 = (unsigned long) tempm;

  RnL = temp1 & 0x0000FFFF;
  RnH = (temp1 >> 16) & 0x0000FFFF;
  RmL = temp2 & 0x0000FFFF;
  RmH = (temp2 >> 16) & 0x0000FFFF;
  
  temp0 = RmL * RnL;
  temp1 = RmH * RnL;
  temp2 = RmL * RnH;
  temp3 = RmH * RnH;

  Res2 = 0;
  Res1 = temp1 + temp2;
  if (Res1 < temp1) Res2 += 0x00010000;

  temp1 = (Res1 << 16) & 0xFFFF0000;
  Res0 = temp0 + temp1;
  if (Res0 < temp0) Res2++;
  
  Res2 = Res2 + ((Res1 >> 16) & 0x0000FFFF) + temp3;

  if (fnLmL < 0) {
    Res2 = ~Res2;
    if (Res0 == 0)
      Res2++;
    else
      Res0 =(~Res0) + 1;
  }
  MACH = Res2;
  MACL = Res0;
  PC += 2;
}

void SuperH::dmulu(void) {
  unsigned long RnL,RnH,RmL,RmH,Res0,Res1,Res2;
  unsigned long temp0,temp1,temp2,temp3;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  RnL = R[n] & 0x0000FFFF;
  RnH = (R[n] >> 16) & 0x0000FFFF;
  RmL = R[m] & 0x0000FFFF;
  RmH = (R[m] >> 16) & 0x0000FFFF;

  temp0 = RmL * RnL;
  temp1 = RmH * RnL;
  temp2 = RmL * RnH;
  temp3 = RmH * RnH;
  
  Res2 = 0;
  Res1 = temp1 + temp2;
  if (Res1 < temp1) Res2 += 0x00010000;
  
  temp1 = (Res1 << 16) & 0xFFFF0000;
  Res0 = temp0 + temp1;
  if (Res0 < temp0) Res2++;
  
  Res2 = Res2 + ((Res1 >> 16) & 0x0000FFFF) + temp3;

  MACH = Res2;
  MACL = Res0;
  PC += 2;
}

void SuperH::dt(void) {
  long n = Instruction::b(instruction);
  R[n]--;
  if (R[n] == 0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC += 2;
}

void SuperH::extsb(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = R[m];
  if ((R[m]&0x00000080)==0) R[n]&=0x000000FF;
  else R[n]|=0xFFFFFF00;
  PC += 2;
}

void SuperH::extsw(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = R[m];
  if ((R[m]&0x00008000)==0) R[n]&=0x0000FFFF;
  else R[n]|=0xFFFF0000;
  PC += 2;
}

void SuperH::extub(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = R[m];
  R[n] &= 0x000000FF;
  PC += 2;
}

void SuperH::extuw(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = R[m];
  R[n] &= 0x0000FFFF;
  PC += 2;
}

void SuperH::jmp(void) {
  unsigned long temp;
  long m = Instruction::b(instruction);

  temp=PC;
  PC = R[m] + 4;
  _delai = temp + 2;
}

void SuperH::jsr(void) {
  long m = Instruction::b(instruction);
  PR = PC;
  PC = R[m] + 4;
  _delai = PR + 2;
}

void SuperH::ldcgbr(void) {
  GBR = R[Instruction::b(instruction)];
  PC += 2;
}

void SuperH::ldcmgbr(void) {
  long m = Instruction::b(instruction);

  GBR = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldcmsr(void) {
  long m = Instruction::b(instruction);

  SR.tout = memoire->getGrosmot(R[m]) & 0x000003F3;
  R[m] += 4;
  PC += 2;
}

void SuperH::ldcmvbr(void) {
  long m = Instruction::b(instruction);

  VBR = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldcsr(void) {
  SR.tout = R[Instruction::b(instruction)]&0x000003F3;
  PC += 2;
}

void SuperH::ldcvbr(void) {
  long m = Instruction::b(instruction);

  VBR = R[m];
  PC += 2;
}

void SuperH::ldsmach(void) {
  MACH = R[Instruction::b(instruction)];
  PC+=2;
}

void SuperH::ldsmacl(void) {
  MACL = R[Instruction::b(instruction)];
  PC += 2;
}

void SuperH::ldsmmach(void) {
  long m = Instruction::b(instruction);
  MACH = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldsmmacl(void) {
  long m = Instruction::b(instruction);
  MACL = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldsmpr(void) {
  long m = Instruction::b(instruction);
  PR = memoire->getGrosmot(R[m]);
  R[m] += 4;
  PC += 2;
}

void SuperH::ldspr(void) {
  PR = R[Instruction::b(instruction)];
  PC += 2;
}

void SuperH::macl(void) {
  unsigned long RnL,RnH,RmL,RmH,Res0,Res1,Res2;
  unsigned long temp0,temp1,temp2,temp3;
  long tempm,tempn,fnLmL;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  
  tempn = (long) memoire->getGrosmot(R[n]);
  R[n] += 4;
  tempm = (long) memoire->getGrosmot(R[m]);
  R[m] += 4;


  if ((long) (tempn^tempm) < 0) fnLmL =- 1;
  else fnLmL = 0;
  if (tempn < 0) tempn = 0 - tempn;
  if (tempm < 0) tempm = 0 - tempm;

  temp1 = (unsigned long) tempn;
  temp2 = (unsigned long) tempm;

  RnL = temp1 & 0x0000FFFF;
  RnH = (temp1 >> 16) & 0x0000FFFF;
  RmL = temp2 & 0x0000FFFF;
  RmH = (temp2 >> 16) & 0x0000FFFF;

  temp0 = RmL * RnL;
  temp1 = RmH * RnL;
  temp2 = RmL * RnH;
  temp3 = RmH * RnH;
     
  Res2 = 0;
  Res1 = temp1 + temp2;
  if (Res1 < temp1) Res2 += 0x00010000;

  temp1 = (Res1 << 16) & 0xFFFF0000;
  Res0 = temp0 + temp1;
  if (Res0 < temp0) Res2++;

  Res2=Res2+((Res1>>16)&0x0000FFFF)+temp3;

  if(fnLmL < 0){
    Res2=~Res2;
    if (Res0==0) Res2++;
    else Res0=(~Res0)+1;
  }
  if(SR.partie.S == 1){
    Res0=MACL+Res0;
    if (MACL>Res0) Res2++;
    Res2+=(MACH&0x0000FFFF);

    if(((long)Res2<0)&&(Res2<0xFFFF8000)){
      Res2=0x00008000;
      Res0=0x00000000;
    }
    if(((long)Res2>0)&&(Res2>0x00007FFF)){
      Res2=0x00007FFF;
      Res0=0xFFFFFFFF;
    };
    
    MACH=Res2;
    MACL=Res0;
  }
  else {
    Res0=MACL+Res0;
    if (MACL>Res0) Res2++;
    Res2+=MACH;

    MACH=Res2;
    MACL=Res0;
  }
  PC+=2;
}

void SuperH::macw(void) {
  long tempm,tempn,dest,src,ans;
  unsigned long templ;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  tempn=(long) memoire->getMot(R[n]);
  R[n]+=2;
  tempm=(long) memoire->getMot(R[m]);
  R[m]+=2;
  templ=MACL;
  tempm=((long)(short)tempn*(long)(short)tempm);

  if ((long)MACL>=0) dest=0;
  else dest=1;
  if ((long)tempm>=0) {
    src=0;
    tempn=0;
  }
  else {
    src=1;
    tempn=0xFFFFFFFF;
  }
  src+=dest;
  MACL+=tempm;
  if ((long)MACL>=0) ans=0;
  else ans=1;
  ans+=dest;
  if (SR.partie.S == 1) {
    if (ans==1) {
      if (src==0) MACL=0x7FFFFFFF;
      if (src==2) MACL=0x80000000;
    }
  }
  else {
    MACH+=tempn;
    if (templ>MACL) MACH+=1;
  }
  PC+=2;
}

void SuperH::mov(void) {
  R[Instruction::b(instruction)]=R[Instruction::c(instruction)];
  PC+=2;
}

void SuperH::mova(void) {
  long disp;
  long d = Instruction::cd(instruction);

  disp=(0x000000FF & (long)d);
  R[0]=(PC&0xFFFFFFFC)+(disp<<2);
  PC+=2;
}

void SuperH::movbl(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = (long) memoire->getOctet(R[m]);
  if ((R[n] & 0x80) == 0)
    R[n] &= 0x000000FF;
  else
    R[n] |= 0xFFFFFF00;
  PC += 2;
}

void SuperH::movbl0(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n]=(long) memoire->getOctet(R[m] + R[0]);
  if ((R[n]&0x80)==0)
    R[n] &= 0x000000FF;
  else
    R[n] |= 0xFFFFFF00;
  PC += 2;
}

void SuperH::movbl4(void) {
  long disp;
  long m = Instruction::c(instruction);
  long d = Instruction::d(instruction);

  disp=(0x0000000F & (long)d);
  R[0] = memoire->getOctet(R[m] + disp);
  if ((R[0]&0x80)==0) R[0]&=0x000000FF;
  else R[0]|=0xFFFFFF00;
  PC+=2;
}

void SuperH::movblg(void) {
  long disp;
  long d = Instruction::cd(instruction);
  
  disp=(0x000000FF & (long)d);
  R[0]=(long) memoire->getOctet(GBR + disp);
  if ((R[0]&0x80)==0) R[0]&=0x000000FF;
  else R[0]|=0xFFFFFF00;
  PC+=2;
}

void SuperH::movbm(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  memoire->setOctet((R[n] - 1),R[m]);
  R[n] -= 1;
  PC += 2;
}

void SuperH::movbp(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = (long) memoire->getOctet(R[m]);
  if ((R[n]&0x80)==0)
    R[n] &= 0x000000FF;
  else
    R[n] |= 0xFFFFFF00;
  if (n != m)
    R[m] += 1;
  PC += 2;
}

void SuperH::movbs(void) {
  memoire->setOctet(R[Instruction::b(instruction)],
	R[Instruction::c(instruction)]);
  PC += 2;
}

void SuperH::movbs0(void) {
  memoire->setOctet(R[Instruction::b(instruction)] + R[0],
	R[Instruction::c(instruction)]);
  PC += 2;
}

void SuperH::movbs4(void) {
  long disp;
  long d = Instruction::d(instruction);
  long n = Instruction::c(instruction);

  disp=(0x0000000F & (long)d);
  memoire->setOctet(R[n]+disp,R[0]);
  PC+=2;
}

void SuperH::movbsg(void) {
  long disp;
  long d = Instruction::cd(instruction);

  disp=(0x000000FF & (long)d);
  memoire->setOctet(GBR + disp,R[0]);
  PC += 2;
}

void SuperH::movi(void) {
  long i = Instruction::cd(instruction);
  long n = Instruction::b(instruction);

  if ((i & 0x80) == 0)
    R[n] = (0x000000FF & (long)i);
  else
    R[n] = (0xFFFFFF00 | (long)i);
  PC += 2;
}

void SuperH::movli(void) {
  long disp;
  long d = Instruction::cd(instruction);
  long n = Instruction::b(instruction);

  disp=(0x000000FF & (long)d);
  R[n] = memoire->getGrosmot((PC & 0xFFFFFFFC) + (disp << 2));
  PC += 2;
}

void SuperH::movll(void) {
  R[Instruction::b(instruction)] =
	memoire->getGrosmot(R[Instruction::c(instruction)]);
  PC += 2;
}

void SuperH::movll0(void) {
  R[Instruction::b(instruction)] =
	memoire->getGrosmot(R[Instruction::c(instruction)] + R[0]);
  PC += 2;
}

void SuperH::movll4(void) {
  long disp;
  long m = Instruction::c(instruction);
  long d = Instruction::d(instruction);
  long n = Instruction::b(instruction);

  disp=(0x0000000F & (long)d);
  R[n] = memoire->getGrosmot(R[m] + (disp << 2));
  PC += 2;
}

void SuperH::movllg(void) {
  long disp;
  long d = Instruction::cd(instruction);

  disp=(0x000000FF & (long)d);
  R[0] = memoire->getGrosmot(GBR + (disp << 2));
  PC+=2;
}

void SuperH::movlm(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  memoire->setGrosmot(R[n] - 4,R[m]);
  R[n] -= 4;
  PC += 2;
}

void SuperH::movlp(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = memoire->getGrosmot(R[m]);
  if (n != m) R[m] += 4;
  PC += 2;
}

void SuperH::movls(void) {
  memoire->setGrosmot(R[Instruction::b(instruction)],
	R[Instruction::c(instruction)]);
  PC += 2;
}

void SuperH::movls0(void) {
  memoire->setGrosmot(R[Instruction::b(instruction)] + R[0],
	R[Instruction::c(instruction)]);
  PC += 2;
}

void SuperH::movls4(void) {
  long disp;
  long m = Instruction::c(instruction);
  long d = Instruction::d(instruction);
  long n = Instruction::b(instruction);

  disp = (0x0000000F & (long) d);
  memoire->setGrosmot(R[n]+(disp<<2),R[m]);
  PC += 2;
}

void SuperH::movlsg(void) {
  long disp;
  long d = Instruction::cd(instruction);

  disp=(0x000000FF & (long)d);
  memoire->setGrosmot(GBR+(disp<<2),R[0]);
  PC+=2;
}

void SuperH::movt(void) {
  R[Instruction::b(instruction)] = (0x00000001 & SR.tout);
  PC += 2;
}

void SuperH::movwi(void) {
  long disp;
  long d = Instruction::cd(instruction);
  long n = Instruction::b(instruction);

  disp=(0x000000FF & (long)d);
  R[n]=(long) memoire->getMot(PC + (disp<<1));
  if ((R[n]&0x8000)==0) R[n]&=0x0000FFFF;
  else R[n]|=0xFFFF0000;
  PC+=2;
}

void SuperH::movwl(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n]=(long) memoire->getMot(R[m]);
  if ((R[n] & 0x8000) == 0)
    R[n] &= 0x0000FFFF;
  else
    R[n] |= 0xFFFF0000;
  PC += 2;
}

void SuperH::movwl0(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n]=(long) memoire->getMot(R[m]+R[0]);
  if ((R[n]&0x8000)==0)
    R[n] &= 0x0000FFFF;
  else
    R[n] |= 0xFFFF0000;
  PC+=2;
}

void SuperH::movwl4(void) {
  long disp;
  long m = Instruction::c(instruction);
  long d = Instruction::d(instruction);

  disp=(0x0000000F & (long)d);
  R[0] = memoire->getMot(R[m]+(disp<<1));
  if ((R[0]&0x8000)==0) R[0]&=0x0000FFFF;
  else R[0]|=0xFFFF0000;
  PC+=2;
}

void SuperH::movwlg(void) {
  long disp;
  long d = Instruction::cd(instruction);

  disp=(0x000000FF & (long)d);
  R[0]=(long) memoire->getMot(GBR+(disp<<1));
  if ((R[0]&0x8000)==0) R[0]&=0x0000FFFF;
  else R[0]|=0xFFFF0000;
  PC += 2;
}

void SuperH::movwm(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  memoire->setMot(R[n] - 2,R[m]);
  R[n] -= 2;
  PC += 2;
}

void SuperH::movwp(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  R[n] = (long) memoire->getMot(R[m]);
  if ((R[n]&0x8000)==0)
    R[n] &= 0x0000FFFF;
  else
    R[n] |= 0xFFFF0000;
  if (n != m)
    R[m] += 2;
  PC += 2;
}

void SuperH::movws(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  memoire->setMot(R[n],R[m]);
  PC += 2;
}

void SuperH::movws0(void) {
  memoire->setMot(R[Instruction::b(instruction)] + R[0],
	R[Instruction::c(instruction)]);
  PC+=2;
}

void SuperH::movws4(void) {
  long disp;
  long d = Instruction::d(instruction);
  long n = Instruction::c(instruction);

  disp=(0x0000000F & (long)d);
  memoire->setMot(R[n]+(disp<<1),R[0]);
  PC+=2;
}

void SuperH::movwsg(void) {
  long disp;
  long d = Instruction::cd(instruction);

  disp=(0x000000FF & (long)d);
  memoire->setMot(GBR+(disp<<1),R[0]);
  PC+=2;
}

void SuperH::mull(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  MACL = R[n] * R[m];
  PC+=2;
}

void SuperH::muls(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  MACL = ((long)(short)R[n]*(long)(short)R[m]);
  PC += 2;
}

void SuperH::mulu(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  MACL = ((unsigned long)(unsigned short)R[n]
	*(unsigned long)(unsigned short)R[m]);
  PC+=2;
}

void SuperH::neg(void) {
  R[Instruction::b(instruction)]=0-R[Instruction::c(instruction)];
  PC+=2;
}

void SuperH::negc(void) {
  unsigned long temp;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  
  temp=0-R[m];
  R[n] = temp - SR.partie.T;
  if (0 < temp) SR.partie.T=1;
  else SR.partie.T=0;
  if (temp < R[n]) SR.partie.T=1;
  PC+=2;
}

void SuperH::nop(void) {
  PC += 2;
}

void SuperH::y_not(void) {
  R[Instruction::b(instruction)] = ~R[Instruction::c(instruction)];
  PC += 2;
}

void SuperH::y_or(void) {
  R[Instruction::b(instruction)] |= R[Instruction::c(instruction)];
  PC += 2;
}

void SuperH::ori(void) {
  R[0] |= (0x000000FF & (long) Instruction::cd(instruction));
  PC += 2;
}

void SuperH::orm(void) {
  long temp;
  long source = Instruction::cd(instruction);

  temp = (long) memoire->getOctet(GBR + R[0]);
  temp |= (0x000000FF & (long) source);
  memoire->setOctet(GBR + R[0],temp);
  PC += 2;
}

void SuperH::rotcl(void) {
  long temp;
  long n = Instruction::b(instruction);

  if ((R[n]&0x80000000)==0) temp=0;
  else temp=1;
  R[n]<<=1;
  if (SR.partie.T == 1) R[n]|=0x00000001;
  else R[n]&=0xFFFFFFFE;
  if (temp==1) SR.partie.T=1;
  else SR.partie.T=0;
  PC+=2;
}

void SuperH::rotcr(void) {
  long temp;
  long n = Instruction::b(instruction);

  if ((R[n]&0x00000001)==0) temp=0;
  else temp=1;
  R[n]>>=1;
  if (SR.partie.T == 1) R[n]|=0x80000000;
  else R[n]&=0x7FFFFFFF;
  if (temp==1) SR.partie.T=1;
  else SR.partie.T=0;
  PC+=2;
}

void SuperH::rotl(void) {
  long n = Instruction::b(instruction);

  if ((R[n]&0x80000000)==0) SR.partie.T=0;
  else SR.partie.T=1;
  R[n]<<=1;
  if (SR.partie.T==1) R[n]|=0x00000001;
  else R[n]&=0xFFFFFFFE;
  PC+=2;
}

void SuperH::rotr(void) {
  long n = Instruction::b(instruction);

  if ((R[n]&0x00000001)==0) SR.partie.T = 0;
  else SR.partie.T = 1;
  R[n]>>=1;
  if (SR.partie.T == 1) R[n]|=0x80000000;
  else R[n]&=0x7FFFFFFF;
  PC+=2;
}

void SuperH::rte() {
  unsigned long temp;

  temp=PC;
  PC = memoire->getGrosmot(R[15]); // + 4;
  R[15] += 4;
  SR.tout = memoire->getGrosmot(R[15]) & 0x000003F3;
  R[15] += 4;
  _delai = temp + 2;
}

void SuperH::rts() {
  unsigned long temp;
  
  temp = PC;
  PC = PR + 4;

  _delai = temp + 2;
}

void SuperH::sett(void) {
  SR.partie.T = 1;
  PC += 2;
}

void SuperH::shal(void) {
  long n = Instruction::b(instruction);
  if ((R[n] & 0x80000000) == 0) SR.partie.T = 0;
  else SR.partie.T = 1;
  R[n] <<= 1;
  PC += 2;
}

void SuperH::shar(void) {
  long temp;
  long n = Instruction::b(instruction);

  if ((R[n]&0x00000001)==0) SR.partie.T = 0;
  else SR.partie.T = 1;
  if ((R[n]&0x80000000)==0) temp = 0;
  else temp = 1;
  R[n] >>= 1;
  if (temp == 1) R[n] |= 0x80000000;
  else R[n] &= 0x7FFFFFFF;
  PC += 2;
}

void SuperH::shll(void) {
  long n = Instruction::b(instruction);

  if ((R[n]&0x80000000)==0) SR.partie.T=0;
  else SR.partie.T=1;
  R[n]<<=1;
  PC+=2;
}

void SuperH::shll2(void) {
  R[Instruction::b(instruction)] <<= 2;
  PC+=2;
}

void SuperH::shll8(void) {
  R[Instruction::b(instruction)]<<=8;
  PC+=2;
}

void SuperH::shll16(void) {
  R[Instruction::b(instruction)]<<=16;
  PC+=2;
}

void SuperH::shlr(void) {
  long n = Instruction::b(instruction);
  if ((R[n]&0x00000001)==0) SR.partie.T=0;
  else SR.partie.T=1;
  R[n]>>=1;
  R[n]&=0x7FFFFFFF;
  PC+=2;
}

void SuperH::shlr2(void) {
  long n = Instruction::b(instruction);
  R[n]>>=2;
  R[n]&=0x3FFFFFFF;
  PC+=2;
}

void SuperH::shlr8(void) {
  long n = Instruction::b(instruction);
  R[n]>>=8;
  R[n]&=0x00FFFFFF;
  PC+=2;
}

void SuperH::shlr16(void) {
  long n = Instruction::b(instruction);
  R[n]>>=16;
  R[n]&=0x0000FFFF;
  PC+=2;
}

void SuperH::stcgbr(void) {
  long n = Instruction::b(instruction);
  R[n]=GBR;
  PC+=2;
}

void SuperH::stcmgbr(void) {
  long n = Instruction::b(instruction);
  R[n]-=4;
  memoire->setGrosmot(R[n],GBR);
  PC+=2;
}

void SuperH::stcmsr(void) {
  long n = Instruction::b(instruction);
  R[n]-=4;
  memoire->setGrosmot(R[n],SR.tout);
  PC+=2;
}

void SuperH::stcmvbr(void) {
  long n = Instruction::b(instruction);
  R[n]-=4;
  memoire->setGrosmot(R[n],VBR);
  PC+=2;
}

void SuperH::stcsr(void) {
  long n = Instruction::b(instruction);
  R[n] = SR.tout;
  PC+=2;
}

void SuperH::stcvbr(void) {
  long n = Instruction::b(instruction);
  R[n]=VBR;
  PC+=2;
}

void SuperH::stsmach(void) {
  long n = Instruction::b(instruction);
  R[n]=MACH;
  PC+=2;
}

void SuperH::stsmacl(void) {
  long n = Instruction::b(instruction);
  R[n]=MACL;
  PC+=2;
}

void SuperH::stsmmach(void) {
  long n = Instruction::b(instruction);
  R[n] -= 4;
  memoire->setGrosmot(R[n],MACH); 
  PC+=2;
}

void SuperH::stsmmacl(void) {
  long n = Instruction::b(instruction);
  R[n] -= 4;
  memoire->setGrosmot(R[n],MACL);
  PC+=2;
}

void SuperH::stsmpr(void) {
  long n = Instruction::b(instruction);
  R[n] -= 4;
  memoire->setGrosmot(R[n],PR);
  PC+=2;
}

void SuperH::stspr(void) {
  long n = Instruction::b(instruction);
  R[n] = PR;
  PC+=2;
}

void SuperH::sub(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  R[n]-=R[m];
  PC+=2;
}

void SuperH::subc(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  unsigned long tmp0,tmp1;
  
  tmp1 = R[n] - R[m];
  tmp0 = R[n];
  R[n] = tmp1 - SR.partie.T;
  if (tmp0 < tmp1) SR.partie.T = 1;
  else SR.partie.T = 0;
  if (tmp1 < R[n]) SR.partie.T = 1;
  PC += 2;
}

void SuperH::subv(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  long dest,src,ans;

  if ((long)R[n]>=0) dest=0;
  else dest=1;
  if ((long)R[m]>=0) src=0;
  else src=1;
  src+=dest;
  R[n]-=R[m];
  if ((long)R[n]>=0) ans=0;
  else ans=1;
  ans+=dest;
  if (src==1) {
    if (ans==1) SR.partie.T=1;
    else SR.partie.T=0;
  }
  else SR.partie.T=0;
  PC+=2;
}

void SuperH::swapb(void) {
  unsigned long temp0,temp1;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  temp0=R[m]&0xffff0000;
  temp1=(R[m]&0x000000ff)<<8;
  R[n]=(R[m]>>8)&0x000000ff;
  R[n]=R[n]|temp1|temp0;
  PC+=2;
}

void SuperH::swapw(void) {
  unsigned long temp;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  temp=(R[m]>>16)&0x0000FFFF;
  R[n]=R[m]<<16;
  R[n]|=temp;
  PC+=2;
}

void SuperH::tas(void) {
  long temp;
  long n = Instruction::b(instruction);

  temp=(long) memoire->getOctet(R[n]);
  if (temp==0) SR.partie.T=1;
  else SR.partie.T=0;
  temp|=0x00000080;
  memoire->setOctet(R[n],temp);
  PC+=2;
}

void SuperH::trapa(void) {
  long imm;
  long i = Instruction::cd(instruction);

  imm=(0x000000FF & i);
  R[15]-=4;
  memoire->setGrosmot(R[15],SR.tout);
  R[15]-=4;
  memoire->setGrosmot(R[15],PC - 2);
  PC = memoire->getGrosmot(VBR+(imm<<2)) + 4;
}

void SuperH::tst(void) {
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);
  if ((R[n]&R[m])==0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC += 2;
}

void SuperH::tsti(void) {
  long temp;
  long i = Instruction::cd(instruction);

  temp=R[0]&(0x000000FF & (long)i);
  if (temp==0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC+=2;
}

void SuperH::tstm(void) {
  long temp;
  long i = Instruction::cd(instruction);

  temp=(long) memoire->getOctet(GBR+R[0]);
  temp&=(0x000000FF & (long)i);
  if (temp==0) SR.partie.T = 1;
  else SR.partie.T = 0;
  PC+=2;
}

void SuperH::y_xor(void) {
  R[Instruction::b(instruction)] ^= R[Instruction::c(instruction)];
  PC += 2;
}

void SuperH::xori(void) {
  long source = Instruction::cd(instruction);
  R[0] ^= (0x000000FF & (long) source);
  PC += 2;
}

void SuperH::xorm(void) {
  long source = Instruction::cd(instruction);
  long temp;

  temp = (long) memoire->getOctet(GBR + R[0]);
  temp ^= (0x000000FF & (long) source);
  memoire->setOctet(GBR + R[0],temp);
  PC += 2;
}

void SuperH::xtrct(void) {
  unsigned long temp;
  long m = Instruction::c(instruction);
  long n = Instruction::b(instruction);

  temp=(R[m]<<16)&0xFFFF0000;
  R[n]=(R[n]>>16)&0x0000FFFF;
  R[n]|=temp;
  PC+=2;
}

/********************************* NON IMPLEMENTE **************/
void SuperH::sleep(void) { throw InstructionNonImplementee("sleep"); }
