/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "superh.hh"

SuperH::opcode SuperH::decode(void) {
  
  switch (Instruction::a(instruction)) {
  case 0:
    switch (Instruction::d(instruction)) {
    case 2:
      switch (Instruction::c(instruction)) {
      case 0: return &SuperH::stcsr;
      case 1: return &SuperH::stcgbr;
      case 2: return &SuperH::stcvbr;
      default: return &SuperH::undecoded;
      }
     
    case 3:
      switch (Instruction::c(instruction)) {
      case 0: return &SuperH::bsrf;
      case 2: return &SuperH::braf;
      default: return &SuperH::undecoded;
      }
     
    case 4: return &SuperH::movbs0;
    case 5: return &SuperH::movws0;
    case 6: return &SuperH::movls0;
    case 7: return &SuperH::mull;
    case 8:
      switch (Instruction::c(instruction)) {
      case 0: return &SuperH::clrt;
      case 1: return &SuperH::sett;
      case 2: return &SuperH::clrmac;
      default: return &SuperH::undecoded;
      }
     
    case 9:
      switch (Instruction::c(instruction)) {
      case 0: return &SuperH::nop;
      case 1: return &SuperH::div0u;
      case 2: return &SuperH::movt;
      default: return &SuperH::undecoded;
      }
     
    case 10:
      switch (Instruction::c(instruction)) {
      case 0: return &SuperH::stsmach;
      case 1: return &SuperH::stsmacl;
      case 2: return &SuperH::stspr;
      default: return &SuperH::undecoded;
      }
     
    case 11:
      switch (Instruction::c(instruction)) {
      case 0: return &SuperH::rts;
      case 1: return &SuperH::sleep;
      case 2: return &SuperH::rte;
      default: return &SuperH::undecoded;
      }
     
    case 12: return &SuperH::movbl0;
    case 13: return &SuperH::movwl0;
    case 14: return &SuperH::movll0;
    case 15: return &SuperH::macl;
    default: return &SuperH::undecoded;
    }
   
  case 1: return &SuperH::movls4;
  case 2:
    switch (Instruction::d(instruction)) {
    case 0: return &SuperH::movbs;
    case 1: return &SuperH::movws;
    case 2: return &SuperH::movls;
    case 4: return &SuperH::movbm;
    case 5: return &SuperH::movwm;
    case 6: return &SuperH::movlm;
    case 7: return &SuperH::div0s;
    case 8: return &SuperH::tst;
    case 9: return &SuperH::y_and;
    case 10: return &SuperH::y_xor;
    case 11: return &SuperH::y_or;
    case 12: return &SuperH::cmpstr;
    case 13: return &SuperH::xtrct;
    case 14: return &SuperH::mulu;
    case 15: return &SuperH::muls;
    default: return &SuperH::undecoded;
    }
   
  case 3:
    switch(Instruction::d(instruction)) {
    case 0:  return &SuperH::cmpeq;
    case 2:  return &SuperH::cmphs;
    case 3:  return &SuperH::cmpge;
    case 4:  return &SuperH::div1;
    case 5:  return &SuperH::dmulu;
    case 6:  return &SuperH::cmphi;
    case 7:  return &SuperH::cmpgt;
    case 8:  return &SuperH::sub;
    case 10: return &SuperH::subc;
    case 11: return &SuperH::subv;
    case 12: return &SuperH::add;
    case 13: return &SuperH::dmuls;
    case 14: return &SuperH::addc;
    case 15: return &SuperH::addv;
    default: return &SuperH::undecoded;
    }
   
  case 4:
    switch(Instruction::d(instruction)) {
    case 0:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::shll;
      case 1: return &SuperH::dt;
      case 2: return &SuperH::shal;
      default: return &SuperH::undecoded;
      }
     
    case 1:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::shlr;
      case 1: return &SuperH::cmppz;
      case 2: return &SuperH::shar;
      default: return &SuperH::undecoded;
      }
     
    case 2:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::stsmmach;
      case 1: return &SuperH::stsmmacl;
      case 2: return &SuperH::stsmpr;
      default: return &SuperH::undecoded;
      }
     
    case 3:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::stcmsr;
      case 1: return &SuperH::stcmgbr;
      case 2: return &SuperH::stcmvbr;
      default: return &SuperH::undecoded;
      }
     
    case 4:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::rotl;
      case 2: return &SuperH::rotcl;
      default: return &SuperH::undecoded;
      }
     
    case 5:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::rotr;
      case 1: return &SuperH::cmppl;
      case 2: return &SuperH::rotcr;
      default: return &SuperH::undecoded;
      }
     
    case 6:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::ldsmmach;
      case 1: return &SuperH::ldsmmacl;
      case 2: return &SuperH::ldsmpr;
      default: return &SuperH::undecoded;
      }
     
    case 7:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::ldcmsr;
      case 1: return &SuperH::ldcmgbr;
      case 2: return &SuperH::ldcmvbr;
      default: return &SuperH::undecoded;
      }
     
    case 8:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::shll2;
      case 1: return &SuperH::shll8;
      case 2: return &SuperH::shll16;
      default: return &SuperH::undecoded;
      }
     
    case 9:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::shlr2;
      case 1: return &SuperH::shlr8;
      case 2: return &SuperH::shlr16;
      default: return &SuperH::undecoded;
      }
     
    case 10:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::ldsmach;
      case 1: return &SuperH::ldsmacl;
      case 2: return &SuperH::ldspr;
      default: return &SuperH::undecoded;
      }
     
    case 11:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::jsr;
      case 1: return &SuperH::tas;
      case 2: return &SuperH::jmp;
      default: return &SuperH::undecoded;
      }
     
    case 14:
      switch(Instruction::c(instruction)) {
      case 0: return &SuperH::ldcsr;
      case 1: return &SuperH::ldcgbr;
      case 2: return &SuperH::ldcvbr;
      default: return &SuperH::undecoded;
      }
     
    case 15: return &SuperH::macw;
    default: return &SuperH::undecoded;
    }
   
  case 5: return &SuperH::movll4;
  case 6:
    switch (Instruction::d(instruction)) {
    case 0:  return &SuperH::movbl;
    case 1:  return &SuperH::movwl;
    case 2:  return &SuperH::movll;
    case 3:  return &SuperH::mov;
    case 4:  return &SuperH::movbp;
    case 5:  return &SuperH::movwp;
    case 6:  return &SuperH::movlp;
    case 7:  return &SuperH::y_not;
    case 8:  return &SuperH::swapb;
    case 9:  return &SuperH::swapw;
    case 10: return &SuperH::negc;
    case 11: return &SuperH::neg;
    case 12: return &SuperH::extub;
    case 13: return &SuperH::extuw;
    case 14: return &SuperH::extsb;
    case 15: return &SuperH::extsw;
    }
   
  case 7: return &SuperH::addi;
  case 8:
    switch (Instruction::b(instruction)) {
    case 0:  return &SuperH::movbs4;
    case 1:  return &SuperH::movws4;
    case 4:  return &SuperH::movbl4;
    case 5:  return &SuperH::movwl4;
    case 8:  return &SuperH::cmpim;
    case 9:  return &SuperH::bt;
    case 11: return &SuperH::bf;
    case 13: return &SuperH::bts;
    case 15: return &SuperH::bfs;
    default: return &SuperH::undecoded;
    }
   
  case 9: return &SuperH::movwi;
  case 10: return &SuperH::bra;
  case 11: return &SuperH::bsr;
  case 12:
    switch(Instruction::b(instruction)) {
    case 0:  return &SuperH::movbsg;
    case 1:  return &SuperH::movwsg;
    case 2:  return &SuperH::movlsg;
    case 3:  return &SuperH::trapa;
    case 4:  return &SuperH::movblg;
    case 5:  return &SuperH::movwlg;
    case 6:  return &SuperH::movllg;
    case 7:  return &SuperH::mova;
    case 8:  return &SuperH::tsti;
    case 9:  return &SuperH::andi;
    case 10: return &SuperH::xori;
    case 11: return &SuperH::ori;
    case 12: return &SuperH::tstm;
    case 13: return &SuperH::andm;
    case 14: return &SuperH::xorm;
    case 15: return &SuperH::orm;
    }
   
  case 13: return &SuperH::movli;
  case 14: return &SuperH::movi;
  default: return &SuperH::undecoded;
  }
}
