/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "saturn.hh"
#include "vdp1.hh"
#include "intc.hh"
#include "exception.hh"

RegistresVdp1::RegistresVdp1(void) : Registre("vdp1", 0x18, VDP1_2) {}

Vdp1::Vdp1(void) {
  memoire = new Memoire(0xC0000, VDP1_1);
  registres = new RegistresVdp1();
  inout = true;
}

Vdp1::~Vdp1(void) {
  delete memoire;
  delete registres;
}

void Vdp1::executer(unsigned long adr) {
  unsigned short commande = memoire->getMot(adr);
  unsigned long suivante;

  while (!(commande & 0x8000)) {

    //cerr << "vdp1 : commande = " << commande << '\n';
    //cerr << "vdp1 :   | mode de saut = " << ((commande & 0x7000) >> 12) << '\n';

    // First, process the command
    if (!(commande & 0x7000)) { // if (!skip)
      switch (commande & 0x001F) {
	case 0: //cerr << "vdp1\t: normal sprite draw" << endl;
	  break;
	case 1: //cerr << "vdp1\t:  scaled sprite draw" << endl;
	  break;
	case 2: // distorted sprite draw
	  break;
	case 4: // polygon draw
	  break;
	case 5: // polyline draw
	  break;
	case 6: // line draw
	  break;
	case 8: // user clipping coordinates
	  break;
	case 9: // system clipping coordinates
	  break;
	case 10: // local coordinate
	  break;
	case 0x10: // draw end
	  break;
	default:
	  //cerr << "vdp1\t: Mauvaise commande : "
	  //     << hex << setw(10) << commande << endl;
	  break;
      }
    }

    // Next, determine where to go next
    switch ((commande & 0x3000) >> 12) {
    case 0: // NEXT, jump to following table
      adr += 0x20;
      break;
    case 1: // ASSIGN, jump to CMDLINK
      adr = memoire->getMot(adr + 2) * 8;
      break;
    case 2: // CALL, call a subroutine
      executer(adr + 0x20);
      adr = memoire->getMot(adr + 2) * 8;
      break;
    case 3: // RETURN, return from subroutine
      return;
    }
    //cerr << "vdp1 :   | table suivante = " << adr << '\n';
    try { commande = memoire->getMot(adr); }
    catch (MauvaisAcces ma) { return; }
  }
  registres->setMot(0x10, 3); // FIXME : on a fini, on place deux bits  1 dans un registre
  //cerr << "vdp1 : draw end\n";
}

/*
void Vdp1::lancer(void) {
  unsigned long PC = 0;
  unsigned long retour;
//    unsigned short pointZoom;
//    unsigned short direction;

  unsigned long suivante;

  // On commence par rcuprer la premire table, qui est forcment une table de commande
  // et au dbut de cette table se trouve la premire commande

  unsigned short commande = memoire->getMot(0);

  cerr << "vdp1 : commande = " << commande << '\n';
  cerr << "vdp1 :  | mode de saut = " << ((commande & 0x3000) >> 12) << '\n';

  // Traitement du mode de saut
  switch ((commande & 0x3000) >> 12) {
  case 0: // on passe  la suivante
    suivante = PC + 0x20;
    break;
  case 1: // on va l o nous dit d'aller CMDLINK
    suivante = memoire->getMot(PC + 2) * 8;
    break;
  case 2: // appel  une sous routine
    retour = PC + 0x20;
    suivante = memoire->getMot(PC + 2) * 8;
    break;
  case 3: // retour  la routine principale
    suivante = retour;
    break;
  }
  cerr << "vdp1 : table suivante = " << suivante << '\n';

//    if (!(commande & 0x4000)) { // ce bit indique si on doit ignorer la table

//      // On rcupre le point de zoom
//      pointZoom = (commande & 0x0F00) >> 8;

//      // Et la direction
//      direction = (commande & 0x0030) >> 4;

//      // On s'occupe de CMDPMOD
//      unsigned short cmdpmod = memoire->getMot(PC + 4);
//      bool mon =  (cmdpmod & 0x8000);
//      bool hss =  (cmdpmod & 0x1000);
//      bool pclp = (cmdpmod & 0x0800);
//      bool clip = (cmdpmod & 0x0400);
//      bool cmod = (cmdpmod & 0x0200);
//      bool mesh = (cmdpmod & 0x0100);
//      bool ecd =  (cmdpmod & 0x0080);
//      bool spd =  (cmdpmod & 0x0040);
//      unsigned char colorMode = (cmdpmod & 0x0038) >> 3;
//      unsigned char colorBits = (cmdpmod & 0x0007);

//      // Puis de la forme
//      unsigned short caractere = memoire->getMot(PC + 8) * 8;
//      // et de sa taille
//      unsigned short tailleX = (memoire->getMot(PC + 10) & 0x3F00) >> 8;
//      unsigned short tailleY = memoire->getMot(PC + 10) & 0x00FF;

//      // Maintenant on rcupre les points
//      unsigned short xa = memoire->getMot(PC + 12) & 0x07FF;
//      unsigned short ya = memoire->getMot(PC + 14) & 0x07FF;
//      unsigned short xb = memoire->getMot(PC + 16) & 0x07FF;
//      unsigned short yb = memoire->getMot(PC + 18) & 0x07FF;
//      unsigned short xc = memoire->getMot(PC + 20) & 0x07FF;
//      unsigned short yc = memoire->getMot(PC + 22) & 0x07FF;
//      unsigned short xd = memoire->getMot(PC + 24) & 0x07FF;
//      unsigned short yd = memoire->getMot(PC + 26) & 0x07FF;
    
//       //et enfin la table de gouraud shading
//      unsigned short gouraud = memoire->getMot(PC + 28) * 8;
    
//      switch ((commande & 0x8000) >> 12) {
//      case 0:
//        switch (commande & 0x000F) {
//        case 0:  normalSpriteDraw();    break;
//        case 1:  scaledSpriteDraw();    break;
//        case 2:  distortedSpriteDraw(); break;
//        case 4:  polygonDraw();         break;
//        case 5:  polylineDraw();        break;
//        case 6:  lineDraw();            break;
//        case 8:  userClipping();        break;
//        case 9:  systemClipping();      break;
//        case 10: localCoordinate();     break;
//        default: throw InstructionIllegale(commande);
//        }
//        break;
//      case 1: drawEnd(); break;
//      }
//    }

}
*/

void Vdp1::normalSpriteDraw(void) {
  cerr << "vdp1\t: normal sprite draw" << endl;
}

void Vdp1::scaledSpriteDraw(void) {
  cerr << "vdp1\t:  scaled sprite draw" << endl;
}

void Vdp1::distortedSpriteDraw(void) {
  cerr << "vdp1\t: distorted sprite draw" << endl;
}

void Vdp1::polygonDraw(void) {
  cerr << "vdp1\t: polygon draw" << endl;
}

void Vdp1::polylineDraw(void) {
  cerr << "vdp1\t: polyline draw" << endl;
}

void Vdp1::lineDraw(void) {
  cerr << "vdp1\t: line draw" << endl;
}

void Vdp1::userClipping(void) {
  cerr << "vdp1\t: user clipping" << endl;
}

void Vdp1::systemClipping(void) {
  cerr << "vdp1\t: system clipping" << endl;
}

void Vdp1::localCoordinate(void) {
  cerr << "vdp1\t: local coordinate" << endl;
}

void Vdp1::drawEnd(void) {
  cerr << "vdp1\t: draw end" << endl;
}
