/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "vdp2.hh"
#include "saturn.hh"
#include "scu.hh"
#include "timer.hh"

Vdp2Registers::Vdp2Registers(Scu *scu) : LoggedMemory("vdp2", 0x120) {
  vdp2 = new Vdp2(scu);
  vdp2Thread = SDL_CreateThread((int (*)(void*)) &Vdp2::lancer, vdp2);
}

Vdp2Registers::~Vdp2Registers(void) {
#if DEBUG
  cerr << "stopping vdp2\n";
#endif
  vdp2->stop();
  SDL_WaitThread(vdp2Thread, NULL);
#if DEBUG
  cerr << "vdp2 stopped\n";
#endif
  delete vdp2;
}

Vdp2 *Vdp2Registers::getVdp2(void) {
  return vdp2;
}

Vdp2::Vdp2(Scu *s) {
  scu = s;
  inout = true;
  _stop = false;
}

void Vdp2::lancer(Vdp2 *vdp2) {
  while(!vdp2->_stop) vdp2->executer();
}

void Vdp2::executer(void) {
	Timer t;
	t.waitVBlankIN();
	scu->sendVBlankIN();
	t.waitVBlankOUT();
	scu->sendVBlankOUT();
}
